############################################################################
# --------------------------------------------------------------------------
# @author James Edmondson <james@koshee.ai>
# --------------------------------------------------------------------------
############################################################################

"""
provides conversion from onnx to tensort engine format
"""

import sys
import os
import argparse

import importlib
migraphx_loader = importlib.util.find_spec('migraphx')

if not migraphx_loader:
  print(f"darknet2any: this script requires an installation with migraphx")
  print(f"  to fix this issue from a local install, use scripts/install_amd.sh")
  print(f"  from pip, try pip install darknet2any[amd]")

  exit(1)

os.environ["PYTHONPATH"] = "/opt/rocm/lib"

import migraphx

from darknet2any.tool.utils import *
from darknet2any.tool.darknet2onnx import *

def parse_args(args):
  """
  parses command line arguments

  Args:
  args (list): list of arguments
  Returns:
  dict: a map of arguments as defined by the parser
  """
  parser = argparse.ArgumentParser(
  description="Converts onnx files to migraphx format for AMD gpus",
  add_help=True
  )
  parser.add_argument('-i','--input','--onnx', action='store',
    dest='input', default=None,
    help='the weights file to convert')
  parser.add_argument('-o','--output','--trt',
    action='store', dest='output', default=None,
    help='the onnx file to create (default=filename.onnx)')
  parser.add_argument('--fp16',
    action='store_true', dest='fp16',
    help='quantize for fp16')
  parser.add_argument('--bf16',
    action='store_true', dest='bf16',
    help='quantize for bf16')
  parser.add_argument('--int8',
    action='store_true', dest='int8',
    help='quantize for int8')
  parser.add_argument('--exhaustive', '--exhaustive-tune',
    action='store_true', dest='exhaustive', default=False,
    help='exhaustively optimize for the target GPU')
  parser.add_argument('--cpu',
    action='store_true', dest='cpu',
    help='target a cpu instead of the gpu')

  return parser.parse_args(args)

def convert(input_file, output_file, convert_options):
  """
  converts onnx to mxr format
  """
  model = migraphx.parse_onnx(input_file)

  target = "gpu"
  if convert_options.cpu:
    target = "cpu"

  if convert_options.int8:
    print("onnx2mxr: quantizing for int8")
    migraphx.quantize_int8(model, t=migraphx.get_target(target))
  elif convert_options.fp16:
    print("onnx2mxr: quantizing for fp16")
    migraphx.quantize_fp16(model)
  elif convert_options.bf16:
    print("onnx2mxr: quantizing for bf16")
    migraphx.quantize_bf16(model)
  else:
    print("onnx2mxr: using default quantization of fp32")

  if convert_options.exhaustive:
    print("onnx2mxr: preparing for exhaustive tuning")


  print(f"onnx2mxr: compiling migraphx for {target}")
  model.compile(t=migraphx.get_target(target),
    exhaustive_tune=convert_options.exhaustive)

  print("mxr parameters:")
  print(f"{model.get_parameter_names()}")

  print("mxr parameter shapes:")
  print(f"{model.get_parameter_shapes()}")

  migraphx.save(model, output_file, format='msgpack')

  
  time.sleep(3.0)

def main():
  """
  main script entry point
  """

  options = parse_args(sys.argv[1:])

  if options.input is not None:
    if not os.path.isfile(options.input):
      print(f"onnx2mxr: onnx file cannot be read. "
        "check file exists or permissions.")
      exit(1)

    prefix = os.path.splitext(options.input)[0]

    input_file = f"{prefix}.onnx"
    output_file = f"{prefix}.mxr"

    if options.output is not None:
      output_file = options.output

    print(f"onnx2mxr: converting onnx to trt...")
    print(f"  source: {input_file}")
    print(f"  target: {output_file}")

    start = time.perf_counter()
    convert(input_file, output_file, options)
    end = time.perf_counter()
    total = end - start

    print("onnx2mxr: conversion complete")

    print(f"onnx2mxr: built {output_file} in {total:.4f}s")

  else:
    parse_args(["-h"])

if __name__ == '__main__':
  main()



############################################################################
# --------------------------------------------------------------------------
# @author James Edmondson <james@koshee.ai>
# --------------------------------------------------------------------------
############################################################################

"""
predicts using tensort on a directory structure of images
"""

import sys
import os
import cv2
import argparse
import numpy as np
import time
import torch
from threading import Thread

import importlib

tensorrt_loader = importlib.util.find_spec('tensorrt')

if not tensorrt_loader:
  print(f"darknet2any: this script requires an installation with tensorrt")
  print(f"  to fix this issue from a local install, use scripts/install_tensorrt.sh")
  print(f"  from pip, try pip install darknet2any[tensorrt]")

  exit(0)

import pycuda.autoinit
import pycuda.driver as cuda
import tensorrt as trt

from darknet2any.tool.utils import plot_boxes_cv2, post_processing, load_class_names

class HostDeviceMem(object):
  def __init__(self, host_mem, device_mem):
    self.host = host_mem
    self.device = device_mem

  def __str__(self):
    return "Host:\n" + str(self.host) + "\nDevice:\n" + str(self.device)

  def __repr__(self):
    return self.__str__()

class Stats:
  """
  a stats class for aggregating thread metrics
  """
  def __init__(self, load_time):
    """
    constructor
    """
    self.fps = 0
    self.embed_time = 0
    self.load_time = load_time
    self.read_time = 0
    self.predict_time = 0
    self.process_time = 0
    self.write_time = 0
    self.context_create_time = 0
    self.avg_embed_time = 0
    self.avg_read_time = 0
    self.avg_predict_time = 0
    self.avg_process_time = 0
    self.avg_write_time = 0
    self.avg_context_create_time = 0
    self.num_predicts = 0
    self.num_threads = 0
  
  def add(self, thread):
    """
    adds thread stats to the aggregated stats
    """
    self.num_predicts += thread.num_predicts
    self.embed_time += thread.embed_time
    self.read_time += thread.read_time
    self.predict_time += thread.predict_time
    self.process_time += thread.process_time
    self.write_time += thread.write_time
    self.context_create_time += thread.context_create_time
    self.num_threads += 1

  def print(self, total_time):
    """
    prints the aggregated stats
    """
    self.fps = self.num_predicts / total_time
    self.avg_embed_time = self.embed_time / self.num_predicts
    self.avg_read_time = self.read_time / self.num_predicts
    self.avg_predict_time = self.predict_time / self.num_predicts
    self.avg_process_time = self.process_time / self.num_predicts
    self.avg_write_time = self.write_time / self.num_predicts
    self.avg_context_create_time = self.context_create_time / self.num_threads
    
    print("Overall stats:")
    print(f"  predicts: {self.num_predicts}, time: {total_time:.3f}s, "
      f"fps: {self.fps:.2f}")
    print(f"  engine_create: time: {self.load_time:.3f}s")
    print(f"  context_create: time: {self.context_create_time:.3f}s, "
      f"avg: {self.avg_context_create_time:.2f}s")
    print(f"  reads: time: {self.read_time:.3f}s, avg: {self.avg_read_time:.4f}s")
    print(f"  predicts: time: {self.predict_time:.4f}s, "
      f"avg: {self.avg_predict_time:.4f}s")
    print(f"  writes: time: {self.write_time:.4f}s, "
      f"avg: {self.avg_write_time:.4f}s")
    print(f"  embedding: time: {self.embed_time:.4f}s, "
      f"avg: {self.avg_embed_time:.4f}s")

class TrtThread(Thread):
  """
  threaded tensorrt implementation
  """
  def __init__(self, idx, engine, cuda_context, classes,
    output, num_threads, images):
    """
    constructor
    """
    
    Thread.__init__(self)
    
    self.avg_embed_time = 0
    self.avg_predict_time = 0
    self.avg_process_time = 0
    self.avg_read_time = 0
    self.avg_write_time = 0
    self.bindings = list()
    self.buffers = None
    self.classes = classes
    self.context_create_time = 0
    self.cuda_context = cuda_context
    self.device = None
    self.embed_time = 0
    self.engine = engine
    self.fps = 0
    self.idx = idx
    self.images = self.get_images(images, num_threads)
    self.inputs = list()
    self.num_predicts = 0
    self.num_threads = num_threads
    self.output_dir = output
    self.outputs = list()
    self.predict_time = 0
    self.process_time = 0
    self.read_time = 0
    self.shape = None
    self.stream = None
    self.trt_context = None
    self.write_time = 0
    self.in_run = False

  def allocate_buffers(self):
    '''
    Allocates all buffers required for a context, i.e., host/device inputs/outputs.
    Should be called in run()
    '''
    self.inputs.clear()
    self.outputs.clear()
    self.bindings.clear()
    self.stream = cuda.Stream()

    for i in range(self.engine.num_io_tensors):
      tensor_name = self.engine.get_tensor_name(i)
      size = trt.volume(self.engine.get_tensor_shape(tensor_name))
      dtype = trt.nptype(self.engine.get_tensor_dtype(tensor_name))

      # Allocate host and device buffers
      host_mem = cuda.pagelocked_empty(size, dtype)
      device_mem = cuda.mem_alloc(host_mem.nbytes)

      # Append the device buffer address to device bindings.
      # When cast to int, it's a linear index into the context's memory
      self.bindings.append(int(device_mem))

      # Append to the appropriate input/output list.
      if self.engine.get_tensor_mode(tensor_name) == trt.TensorIOMode.INPUT:
        self.inputs.append(HostDeviceMem(host_mem, device_mem))
      else:
        self.outputs.append(HostDeviceMem(host_mem, device_mem))
        
    self.buffers = self.inputs, self.outputs, self.bindings, self.stream
    return self.buffers

  def close_contexts(self):
    """
    closes all contexts
    """
    self.inputs.clear()
    self.outputs.clear()
    self.bindings.clear()
    
    del self.stream
    del self.trt_context
    
    self.cuda_context.pop()
  
  # This function is generalized for multiple inputs/outputs.
  # inputs and outputs are expected to be lists of HostDeviceMem objects.
  def do_inference(self):
    """
    predicts classes from an engine
    """
    # Setup tensor address
    
    [cuda.memcpy_htod_async(inp.device, inp.host, self.stream) for inp in self.inputs]

    for i in range(self.engine.num_io_tensors):
      self.trt_context.set_tensor_address(self.engine.get_tensor_name(i), self.bindings[i])

    # Run inference
    self.trt_context.execute_async_v3(stream_handle=self.stream.handle)

    [cuda.memcpy_dtoh_async(out.host, out.device, self.stream) for out in self.outputs]

    # Synchronize the stream
    self.stream.synchronize()
    
    # Return only the host outputs.
    return [out.host for out in self.outputs]

  def get_contexts(self):
    """
    retrieves a thread context from an engine. Has to be called within run
    """
    #cuda.init()
    #if self.in_run:
    self.cuda_context.push()
    self.trt_context = self.engine.create_execution_context()
   
    shape = None

    for i in range(self.engine.num_io_tensors):
      tensor_name = self.engine.get_tensor_name(i)
      print(f"  tensor_name[{i}] = {tensor_name}")
      print(f"  tensor_name[{i}].shape = {get_binding_shape(self.engine,tensor_name)}")

    shape = get_binding_shape(self.engine, "input")

    if shape is not None:

      self.shape = (
        shape[3],
        shape[2]
      )
   
  def get_images(self, images, num_threads):
    '''
    Allocates an images list based on num threads and length of images
    '''
    segment = int(len(images) / num_threads)
    start = int(segment * self.idx)
    end = int(segment) * (self.idx + 1)
    
    results = None
    
    if self.idx != num_threads - 1:
      results = images[start:end]
    else:
      results = images[start:]
    
    #print(f"get_images: images={results}")
    
    return results
  
  def predict(self, image_file):
    """
    predicts classes of an image file

    Args:
    engine (tensorrt.ICudaEngine): the trt engine
    context (tensorrt.IExecutionContext]): the trt context for this thread
    buffers (tuple): result of allocate_buffers call
    shape (str): the shape of the engine input (what we're resizing to)
    classes (list): a list of class names (strings)
    output (str): directory to store labeled images to (None means don't save)
    image_file (str): path to an image
    Returns:
    tuple: read_time, predict_time, process_time, write_time
    """
    print(f"trt: Reading {image_file}")

    basename = os.path.basename(image_file)
    start = time.perf_counter()

    img = cv2.imread(image_file)
    resized = cv2.resize(img, self.shape, interpolation=cv2.INTER_LINEAR)
    img_in = cv2.cvtColor(resized, cv2.COLOR_BGR2RGB)
    img_in = np.transpose(img_in, (2, 0, 1)).astype(np.float32)
    img_in = np.expand_dims(img_in, axis=0)
    img_in /= 255.0
    img_in = np.ascontiguousarray(img_in)
    
    end = time.perf_counter()
    read_time = end - start

    start = time.perf_counter()
    # Allocate buffers

    # print('Length of inputs: ', len(inputs))
    self.inputs[0].host = img_in

    trt_outputs = self.do_inference()

    trt_outputs[0] = trt_outputs[0].reshape(1, -1, 1, 4)
    trt_outputs[1] = trt_outputs[1].reshape(1, -1, len(self.classes))
    embeddings = None
    if len(trt_outputs) > 2:
      embeddings = trt_outputs[2]
    embedding_path = f"{self.output_dir}/{basename}.features"

    end = time.perf_counter()
    predict_time = end - start

    start = time.perf_counter()
    boxes = post_processing(img, 0.4, 0.6, trt_outputs)
    end = time.perf_counter()
    process_time = end - start

    start = time.perf_counter()
    if self.output_dir is not None:
      torch.save(embeddings, embedding_path)
    end = time.perf_counter()
    embed_time = end - start
    
    start = time.perf_counter()
    if self.output_dir is not None:
      plot_boxes_cv2(img, boxes[0],
        savename=f"{self.output_dir}/{basename}", class_names=self.classes)
    write_time = time.perf_counter() - start

    print(f"trt: predict for {image_file}")
    print(f"  output: {boxes}")
    print(f"  read_time: {read_time:.4f}s")
    print(f"  predict_time: {predict_time:.4f}s")
    print(f"  post_processing: {process_time:.4f}s")
    print(f"  write_time: {write_time:.4f}s")
    print(f"  embed_time: {embed_time:.4f}s")

    return read_time, predict_time, process_time, write_time, embed_time

  def run(self):
    """
    executes thread logic
    """
    
    self.in_run = True
    if len(self.images) > 0:
      
      start = time.perf_counter()
      self.get_contexts()
      
      self.allocate_buffers()
      
      self.context_create_time = time.perf_counter() - start
      
      start = time.perf_counter()
      for image in self.images:

        print(f"run: predicting on {image}")
        results = self.predict(image)
        read_time, predict_time, process_time, write_time, embed_time = results
        
        self.read_time += read_time
        self.predict_time += predict_time
        self.process_time += process_time
        self.write_time += write_time
        self.embed_time += embed_time
        self.num_predicts += 1
        
      total_time = time.perf_counter() - start
      self.fps = self.num_predicts / total_time
          
      self.avg_embed_time = self.embed_time / self.num_predicts
      self.avg_read_time = self.read_time / self.num_predicts
      self.avg_predict_time = self.predict_time / self.num_predicts
      self.avg_process_time = self.process_time / self.num_predicts
      self.avg_write_time = self.write_time / self.num_predicts

      print(f"thread.{self.idx}: time for {self.num_predicts} predicts")
      print(f"  total_time: {total_time:.4f}s, fps={self.fps:.2f}")
      print(f"  context_create: total: {self.context_create_time:.4f}s")
      print(f"  image_read_time: total: {self.read_time:.4f}s, avg: {self.avg_read_time:.4f}s")
      print(f"  predict_time: {self.predict_time:.4f}s, avg: {self.avg_predict_time:.4f}s")
      print(f"  process_time: {self.process_time:.4f}s, avg: {self.avg_process_time:.4f}s")
      print(f"  write_time: {self.write_time:.4f}s, avg: {self.avg_write_time:.4f}s")
      print(f"  embed_time: {self.embed_time:.4f}s, avg: {self.avg_embed_time:.4f}s")
      
      self.close_contexts()
    else:
      print(f"thread.{self.idx}: no images, so no predicts")
      
    self.in_run = False

def is_image(filename):
  """
  checks if filename is an image

  Args:
  filename (str): a filename to check extensions of
  Returns:
  bool: a map of arguments as defined by the parser
  """
  ext = os.path.splitext(filename)[1].lower()

  return ext == ".jpg" or ext == ".png"

def parse_args(args):
  """
  parses command line arguments

  Args:
  args (list): list of arguments
  Returns:
  dict: a map of arguments as defined by the parser
  """
  parser = argparse.ArgumentParser(
  description="predicts from a trt model",
  add_help=True
  )
  parser.add_argument('-i','--input','--trt', action='store',
    dest='input', default=None,
    help='the engine/trt to load')
  parser.add_argument('--image', action='store',
    dest='image', default=None,
    help='the image to test the model on')
  parser.add_argument('--image-dir', action='store',
    dest='image_dir', default=None,
    help='a directory of images to test')
  parser.add_argument('--no','--no-output', action='store_true',
    dest='no_output', default=False, 
    help='do not write labeled files to disk (performance test)')
  parser.add_argument('-o','--output-dir', action='store',
    dest='output', default="labeled_images",
    help='a directory to place labeled images')
  parser.add_argument('-t','--threads', action='store',
    dest='threads', default=1, type=int,
    help='the number of threads to run')
  
  return parser.parse_args(args)

def load_engine(path):
  """
  loads the binary trt engine file
  """
  TRT_LOGGER = trt.Logger(trt.Logger.WARNING)
  runtime = trt.Runtime(TRT_LOGGER)
  engine = None

  with open(path, "rb") as f:
    engine = runtime.deserialize_cuda_engine(f.read())
  
  return engine

def get_binding_shape(engine, name):
  """
  gets the shape of the image
  """
  return list(engine.get_tensor_shape(name))

def get_binding_dtype(engine, name):
  """
  gets the binding for the data type
  """
  return trt.nptype(engine.get_tensor_dtype(name))

def main():
  """
  main script entry point
  """

  options = parse_args(sys.argv[1:])
  has_images = options.image is not None or options.image_dir is not None
  
  if options.no_output:
    print(f"trt: not saving labeled files")
    options.output = None
  else:
    print(f"trt: saving labeled files to {options.output}")

  print(f"trt: predicting with {options.input}")

  basename, ext = os.path.splitext(options.input)
  if ext == "":
    options.input += ".trt"

  if options.input is not None and has_images:

    if not os.path.isfile(options.input):
      print(f"predict_trt: trt file cannot be read. "
        "check file exists or permissions.")
      exit(1)

    print(f"trt: loading {options.input}")

    basename = os.path.splitext(options.input)[0]

    if options.output is not None and not os.path.isdir(options.output):
      os.makedirs(options.output)

    names_file = f"{basename}.names"
    classes = load_class_names(names_file)
    
    cuda.init()
    
    cuda_context = pycuda.autoinit.context

    # 1. Load the trt model
    start = time.perf_counter()

    engine = load_engine(options.input)
    
    end = time.perf_counter()
    load_time = end - start
    print(f"  load_time: {load_time:.4f}s")

    images = []

    if options.image is not None:
      images.append(options.image)

    if options.image_dir is not None:

      for dir, _, files in os.walk(options.image_dir):
        for file in files:
          source = f"{dir}/{file}"

          # file needs to be video extension and not already in cameras
          if is_image(file):
            images.append(source)
    
    worker_threads = list()
    stats = Stats(load_time)
    
    for i in range(options.threads):
      print(f"Creating thread.{i}")
      thread = TrtThread(i, engine, cuda_context, classes, options.output,
        options.threads, images)
      worker_threads.append(thread)
    
    work_start = time.perf_counter()
    
    for worker in worker_threads:
      worker.start()
      
    for worker in worker_threads:
      worker.join()
      stats.add(worker)
      
    work_time = time.perf_counter() - work_start
    
    stats.print(work_time)
    
    del engine

  else:

    print("No model or image specified. Printing usage and help.")
    parse_args(["-h"])

if __name__ == '__main__':
  main()



import sys
import os
import time
import math
import torch
import torch.nn as nn
import cv2
import numpy as np
from torch.autograd import Variable

import itertools
import struct  # get_image_size
import imghdr  # get_image_size

from darknet2any.tool import utils 


def bbox_ious(boxes1, boxes2, x1y1x2y2=True):
  if x1y1x2y2:
    mx = torch.min(boxes1[0], boxes2[0])
    Mx = torch.max(boxes1[2], boxes2[2])
    my = torch.min(boxes1[1], boxes2[1])
    My = torch.max(boxes1[3], boxes2[3])
    w1 = boxes1[2] - boxes1[0]
    h1 = boxes1[3] - boxes1[1]
    w2 = boxes2[2] - boxes2[0]
    h2 = boxes2[3] - boxes2[1]
  else:
    mx = torch.min(boxes1[0] - boxes1[2] / 2.0, boxes2[0] - boxes2[2] / 2.0)
    Mx = torch.max(boxes1[0] + boxes1[2] / 2.0, boxes2[0] + boxes2[2] / 2.0)
    my = torch.min(boxes1[1] - boxes1[3] / 2.0, boxes2[1] - boxes2[3] / 2.0)
    My = torch.max(boxes1[1] + boxes1[3] / 2.0, boxes2[1] + boxes2[3] / 2.0)
    w1 = boxes1[2]
    h1 = boxes1[3]
    w2 = boxes2[2]
    h2 = boxes2[3]
    uw = Mx - mx
    uh = My - my
    cw = w1 + w2 - uw
    ch = h1 + h2 - uh
    mask = ((cw <= 0) + (ch <= 0) > 0)
    area1 = w1 * h1
    area2 = w2 * h2
    carea = cw * ch
    carea[mask] = 0
    uarea = area1 + area2 - carea
  return carea / uarea


def get_region_boxes(boxes_and_confs, include_embeddings=False):

  # print('Getting boxes from boxes and confs ...')

  boxes_list = []
  confs_list = []
  features_list = []

  #print(f"get_region_boxes:boxes_and_confs:size={len(boxes_and_confs)}")
  
  for item in boxes_and_confs:
    boxes_list.append(item[0])
    confs_list.append(item[1])
    if include_embeddings:
      features_list.append(item[2])

  # boxes: [batch, num1 + num2 + num3, 1, 4]
  # confs: [batch, num1 + num2 + num3, num_classes]
  boxes = torch.cat(boxes_list, dim=1)
  confs = torch.cat(confs_list, dim=1)
  
  #print(f"get_region_boxes:features_list:shape={features_list[0].shape}")
  
  result = [boxes, confs]
  
  if include_embeddings:
    result.append(features_list[0])
  
  return result


def convert2cpu(gpu_matrix):
  return torch.FloatTensor(gpu_matrix.size()).copy_(gpu_matrix)


def convert2cpu_long(gpu_matrix):
  return torch.LongTensor(gpu_matrix.size()).copy_(gpu_matrix)



def do_detect(model, img, conf_thresh, nms_thresh, use_cuda=1, embedding_path=None, debug=False):
  model.eval()
  with torch.no_grad():
    img_prep_start = time.perf_counter()

    if type(img) == np.ndarray and len(img.shape) == 3:  # cv2 image
      img = torch.from_numpy(img.transpose(2, 0, 1)).float().div(255.0).unsqueeze(0)
    elif type(img) == np.ndarray and len(img.shape) == 4:
      img = torch.from_numpy(img.transpose(0, 3, 1, 2)).float().div(255.0)
    else:
      print("unknow image type")
      exit(-1)

    if use_cuda:
      img = img.cuda()

    img = torch.autograd.Variable(img)

    predict_start = time.perf_counter()

    output = model(img)

    embed_start = time.perf_counter()

    embedding = None
    if len(output) > 2:
      embedding = output[2]

    embed_end = time.perf_counter()

    if embedding_path is not None:
      torch.save(embedding, embedding_path)

    save_embedding_end = time.perf_counter()

    preprocess_time = predict_start - img_prep_start
    predict_time = embed_start - predict_start
    embed_time = embed_end - embed_start
    save_embedding_time = save_embedding_end - embed_end

    if debug:
      print(f"torch: embedding for img is shaped {embedding.shape}")

      print(f"torch: time for do_detect")
      print(f"  preprocess_time: {preprocess_time:.4f}s")
      print(f"  predict_time: {predict_time:.4f}s")
      print(f"  embed_time: {embed_time:.5f}s")
      print(f"  save_embedding_time: {save_embedding_time:.4f}s")

    return utils.post_processing(img, conf_thresh, nms_thresh, output)



import torch
from darknet2any.tool.torch_utils import convert2cpu


def parse_cfg(cfgfile):
  blocks = []
  fp = open(cfgfile, 'r')
  block = None
  line = fp.readline()
  while line != '':
    line = line.rstrip()
    if line == '' or line[0] == '#':
      line = fp.readline()
      continue
    elif line[0] == '[':
      if block:
        blocks.append(block)
      block = dict()
      block['type'] = line.lstrip('[').rstrip(']')
      # set default value
      if block['type'] == 'convolutional':
        block['batch_normalize'] = 0
    else:
      key, value = line.split('=')
      key = key.strip()
      if key == 'type':
        key = '_type'
      value = value.strip()
      block[key] = value
    line = fp.readline()

  if block:
    blocks.append(block)
  fp.close()
  return blocks


def print_cfg(blocks):
  print('layer   filters  size        input        output');
  prev_width = 416
  prev_height = 416
  prev_filters = 3
  out_filters = []
  out_widths = []
  out_heights = []
  ind = -2
  for block in blocks:
    ind = ind + 1
    if block['type'] == 'net':
      prev_width = int(block['width'])
      prev_height = int(block['height'])
      continue
    elif block['type'] == 'convolutional':
      filters = int(block['filters'])
      kernel_size = int(block['size'])
      stride = int(block['stride'])
      is_pad = int(block['pad'])
      pad = (kernel_size - 1) // 2 if is_pad else 0
      width = (prev_width + 2 * pad - kernel_size) // stride + 1
      height = (prev_height + 2 * pad - kernel_size) // stride + 1
      print('%5d %-6s %4d  %d x %d / %d   %3d x %3d x%4d   ->   %3d x %3d x%4d' % (
        ind, 'conv', filters, kernel_size, kernel_size, stride, prev_width, prev_height, prev_filters, width,
        height, filters))
      prev_width = width
      prev_height = height
      prev_filters = filters
      out_widths.append(prev_width)
      out_heights.append(prev_height)
      out_filters.append(prev_filters)
    elif block['type'] == 'maxpool':
      pool_size = int(block['size'])
      stride = int(block['stride'])
      width = prev_width // stride
      height = prev_height // stride
      print('%5d %-6s     %d x %d / %d   %3d x %3d x%4d   ->   %3d x %3d x%4d' % (
        ind, 'max', pool_size, pool_size, stride, prev_width, prev_height, prev_filters, width, height,
        filters))
      prev_width = width
      prev_height = height
      prev_filters = filters
      out_widths.append(prev_width)
      out_heights.append(prev_height)
      out_filters.append(prev_filters)
    elif block['type'] == 'avgpool':
      width = 1
      height = 1
      print('%5d %-6s           %3d x %3d x%4d   ->  %3d' % (
        ind, 'avg', prev_width, prev_height, prev_filters, prev_filters))
      prev_width = width
      prev_height = height
      prev_filters = filters
      out_widths.append(prev_width)
      out_heights.append(prev_height)
      out_filters.append(prev_filters)
    elif block['type'] == 'softmax':
      print('%5d %-6s                  ->  %3d' % (ind, 'softmax', prev_filters))
      out_widths.append(prev_width)
      out_heights.append(prev_height)
      out_filters.append(prev_filters)
    elif block['type'] == 'cost':
      print('%5d %-6s                   ->  %3d' % (ind, 'cost', prev_filters))
      out_widths.append(prev_width)
      out_heights.append(prev_height)
      out_filters.append(prev_filters)
    elif block['type'] == 'reorg':
      stride = int(block['stride'])
      filters = stride * stride * prev_filters
      width = prev_width // stride
      height = prev_height // stride
      print('%5d %-6s       / %d   %3d x %3d x%4d   ->   %3d x %3d x%4d' % (
        ind, 'reorg', stride, prev_width, prev_height, prev_filters, width, height, filters))
      prev_width = width
      prev_height = height
      prev_filters = filters
      out_widths.append(prev_width)
      out_heights.append(prev_height)
      out_filters.append(prev_filters)
    elif block['type'] == 'upsample':
      stride = int(block['stride'])
      filters = prev_filters
      width = prev_width * stride
      height = prev_height * stride
      print('%5d %-6s       * %d   %3d x %3d x%4d   ->   %3d x %3d x%4d' % (
        ind, 'upsample', stride, prev_width, prev_height, prev_filters, width, height, filters))
      prev_width = width
      prev_height = height
      prev_filters = filters
      out_widths.append(prev_width)
      out_heights.append(prev_height)
      out_filters.append(prev_filters)
    elif block['type'] == 'route':
      layers = block['layers'].split(',')
      layers = [int(i) if int(i) > 0 else int(i) + ind for i in layers]
      if len(layers) == 1:
        print('%5d %-6s %d' % (ind, 'route', layers[0]))
        prev_width = out_widths[layers[0]]
        prev_height = out_heights[layers[0]]
        prev_filters = out_filters[layers[0]]
      elif len(layers) == 2:
        print('%5d %-6s %d %d' % (ind, 'route', layers[0], layers[1]))
        prev_width = out_widths[layers[0]]
        prev_height = out_heights[layers[0]]
        assert (prev_width == out_widths[layers[1]])
        assert (prev_height == out_heights[layers[1]])
        prev_filters = out_filters[layers[0]] + out_filters[layers[1]]
      elif len(layers) == 4:
        print('%5d %-6s %d %d %d %d' % (ind, 'route', layers[0], layers[1], layers[2], layers[3]))
        prev_width = out_widths[layers[0]]
        prev_height = out_heights[layers[0]]
        assert (prev_width == out_widths[layers[1]] == out_widths[layers[2]] == out_widths[layers[3]])
        assert (prev_height == out_heights[layers[1]] == out_heights[layers[2]] == out_heights[layers[3]])
        prev_filters = out_filters[layers[0]] + out_filters[layers[1]] + out_filters[layers[2]] + out_filters[
          layers[3]]
      else:
        print("route error !!! {} {} {}".format(sys._getframe().f_code.co_filename,
                            sys._getframe().f_code.co_name, sys._getframe().f_lineno))

      out_widths.append(prev_width)
      out_heights.append(prev_height)
      out_filters.append(prev_filters)
    elif block['type'] in ['region', 'yolo']:
      print('%5d %-6s' % (ind, 'detection'))
      out_widths.append(prev_width)
      out_heights.append(prev_height)
      out_filters.append(prev_filters)
    elif block['type'] == 'shortcut':
      from_id = int(block['from'])
      from_id = from_id if from_id > 0 else from_id + ind
      print('%5d %-6s %d' % (ind, 'shortcut', from_id))
      prev_width = out_widths[from_id]
      prev_height = out_heights[from_id]
      prev_filters = out_filters[from_id]
      out_widths.append(prev_width)
      out_heights.append(prev_height)
      out_filters.append(prev_filters)
    elif block['type'] == 'sam':
      from_id = int(block['from'])
      from_id = from_id if from_id > 0 else from_id + ind
      print('%5d %-6s %d' % (ind, 'sam', from_id))
      prev_width = out_widths[from_id]
      prev_height = out_heights[from_id]
      prev_filters = out_filters[from_id]
      out_widths.append(prev_width)
      out_heights.append(prev_height)
      out_filters.append(prev_filters)
    elif block['type'] == 'connected':
      filters = int(block['output'])
      print('%5d %-6s              %d  ->  %3d' % (ind, 'connected', prev_filters, filters))
      prev_filters = filters
      out_widths.append(1)
      out_heights.append(1)
      out_filters.append(prev_filters)
    else:
      print('unknown type %s' % (block['type']))


def load_conv(buf, start, conv_model):
  num_w = conv_model.weight.numel()
  num_b = conv_model.bias.numel()
  conv_model.bias.data.copy_(torch.from_numpy(buf[start:start + num_b]));
  start = start + num_b
  conv_model.weight.data.copy_(torch.from_numpy(buf[start:start + num_w]).reshape(conv_model.weight.data.shape));
  start = start + num_w
  return start


def save_conv(fp, conv_model):
  if conv_model.bias.is_cuda:
    convert2cpu(conv_model.bias.data).numpy().tofile(fp)
    convert2cpu(conv_model.weight.data).numpy().tofile(fp)
  else:
    conv_model.bias.data.numpy().tofile(fp)
    conv_model.weight.data.numpy().tofile(fp)


def load_conv_bn(buf, start, conv_model, bn_model):
  num_w = conv_model.weight.numel()
  num_b = bn_model.bias.numel()
  bn_model.bias.data.copy_(torch.from_numpy(buf[start:start + num_b]));
  start = start + num_b
  bn_model.weight.data.copy_(torch.from_numpy(buf[start:start + num_b]));
  start = start + num_b
  bn_model.running_mean.copy_(torch.from_numpy(buf[start:start + num_b]));
  start = start + num_b
  bn_model.running_var.copy_(torch.from_numpy(buf[start:start + num_b]));
  start = start + num_b
  conv_model.weight.data.copy_(torch.from_numpy(buf[start:start + num_w]).reshape(conv_model.weight.data.shape));
  start = start + num_w
  return start


def save_conv_bn(fp, conv_model, bn_model):
  if bn_model.bias.is_cuda:
    convert2cpu(bn_model.bias.data).numpy().tofile(fp)
    convert2cpu(bn_model.weight.data).numpy().tofile(fp)
    convert2cpu(bn_model.running_mean).numpy().tofile(fp)
    convert2cpu(bn_model.running_var).numpy().tofile(fp)
    convert2cpu(conv_model.weight.data).numpy().tofile(fp)
  else:
    bn_model.bias.data.numpy().tofile(fp)
    bn_model.weight.data.numpy().tofile(fp)
    bn_model.running_mean.numpy().tofile(fp)
    bn_model.running_var.numpy().tofile(fp)
    conv_model.weight.data.numpy().tofile(fp)


def load_fc(buf, start, fc_model):
  num_w = fc_model.weight.numel()
  num_b = fc_model.bias.numel()
  fc_model.bias.data.copy_(torch.from_numpy(buf[start:start + num_b]));
  start = start + num_b
  fc_model.weight.data.copy_(torch.from_numpy(buf[start:start + num_w]));
  start = start + num_w
  return start


def save_fc(fp, fc_model):
  fc_model.bias.data.numpy().tofile(fp)
  fc_model.weight.data.numpy().tofile(fp)


if __name__ == '__main__':
  import sys

  blocks = parse_cfg('cfg/yolo.cfg')
  if len(sys.argv) == 2:
    blocks = parse_cfg(sys.argv[1])
  print_cfg(blocks)



# -*- coding: utf-8 -*-
'''
@Time      : 2020/04/26 15:48
@Author    : Tianxiaomo
@File      : camera.py
@Noice     :
@Modificattion :
  @Author  :
  @Time    :
  @Detail  :

'''
from __future__ import division
import cv2
from darknet2any.tool.darknet2pytorch import Darknet
import argparse
from darknet2any.tool.utils import *
from darknet2any.tool.torch_utils import *


def arg_parse():
  """
  Parse arguements to the detect module

  """

  parser = argparse.ArgumentParser(description='YOLO v3 Cam Demo')
  parser.add_argument("--confidence", dest="confidence", help="Object Confidence to filter predictions", default=0.25)
  parser.add_argument("--nms_thresh", dest="nms_thresh", help="NMS Threshhold", default=0.4)
  parser.add_argument("--reso", dest='reso', help=
  "Input resolution of the network. Increase to increase accuracy. Decrease to increase speed",
            default="160", type=str)
  return parser.parse_args()


if __name__ == '__main__':
  cfgfile = "cfg/yolov4.cfg"
  weightsfile = "weight/yolov4.weights"

  args = arg_parse()
  confidence = float(args.confidence)
  nms_thesh = float(args.nms_thresh)
  CUDA = torch.cuda.is_available()
  num_classes = 80
  bbox_attrs = 5 + num_classes
  class_names = load_class_names("data/coco.names")

  model = Darknet(cfgfile)
  model.load_weights(weightsfile)

  if CUDA:
    model.cuda()

  model.eval()
  cap = cv2.VideoCapture(0)

  assert cap.isOpened(), 'Cannot capture source'

  frames = 0
  start = time.time()
  while cap.isOpened():
    ret, frame = cap.read()
    if ret:
      sized = cv2.resize(frame, (model.width, model.height))
      sized = cv2.cvtColor(sized, cv2.COLOR_BGR2RGB)
      boxes = do_detect(model, sized, 0.5, 0.4, CUDA)

      orig_im = plot_boxes_cv2(frame, boxes, class_names=class_names)

      cv2.imshow("frame", orig_im)
      key = cv2.waitKey(1)
      if key & 0xFF == ord('q'):
        break
      frames += 1
      print("FPS of the video is {:5.2f}".format(frames / (time.time() - start)))
    else:
      break



import torch.nn as nn
import torch.nn.functional as F
from darknet2any.tool.torch_utils import *


def build_targets(pred_boxes, target, anchors, num_anchors, num_classes, nH, nW, noobject_scale, object_scale,
          sil_thresh, seen):
  nB = target.size(0)
  nA = num_anchors
  nC = num_classes
  anchor_step = len(anchors) / num_anchors
  conf_mask = torch.ones(nB, nA, nH, nW) * noobject_scale
  coord_mask = torch.zeros(nB, nA, nH, nW)
  cls_mask = torch.zeros(nB, nA, nH, nW)
  tx = torch.zeros(nB, nA, nH, nW)
  ty = torch.zeros(nB, nA, nH, nW)
  tw = torch.zeros(nB, nA, nH, nW)
  th = torch.zeros(nB, nA, nH, nW)
  tconf = torch.zeros(nB, nA, nH, nW)
  tcls = torch.zeros(nB, nA, nH, nW)

  nAnchors = nA * nH * nW
  nPixels = nH * nW
  for b in range(nB):
    cur_pred_boxes = pred_boxes[b * nAnchors:(b + 1) * nAnchors].t()
    cur_ious = torch.zeros(nAnchors)
    for t in range(50):
      if target[b][t * 5 + 1] == 0:
        break
      gx = target[b][t * 5 + 1] * nW
      gy = target[b][t * 5 + 2] * nH
      gw = target[b][t * 5 + 3] * nW
      gh = target[b][t * 5 + 4] * nH
      cur_gt_boxes = torch.FloatTensor([gx, gy, gw, gh]).repeat(nAnchors, 1).t()
      cur_ious = torch.max(cur_ious, bbox_ious(cur_pred_boxes, cur_gt_boxes, x1y1x2y2=False))
    conf_mask[b][cur_ious > sil_thresh] = 0
  if seen < 12800:
    if anchor_step == 4:
      tx = torch.FloatTensor(anchors).view(nA, anchor_step).index_select(1, torch.LongTensor([2])).view(1, nA, 1,
                                                        1).repeat(
        nB, 1, nH, nW)
      ty = torch.FloatTensor(anchors).view(num_anchors, anchor_step).index_select(1, torch.LongTensor([2])).view(
        1, nA, 1, 1).repeat(nB, 1, nH, nW)
    else:
      tx.fill_(0.5)
      ty.fill_(0.5)
    tw.zero_()
    th.zero_()
    coord_mask.fill_(1)

  nGT = 0
  nCorrect = 0
  for b in range(nB):
    for t in range(50):
      if target[b][t * 5 + 1] == 0:
        break
      nGT = nGT + 1
      best_iou = 0.0
      best_n = -1
      min_dist = 10000
      gx = target[b][t * 5 + 1] * nW
      gy = target[b][t * 5 + 2] * nH
      gi = int(gx)
      gj = int(gy)
      gw = target[b][t * 5 + 3] * nW
      gh = target[b][t * 5 + 4] * nH
      gt_box = [0, 0, gw, gh]
      for n in range(nA):
        aw = anchors[anchor_step * n]
        ah = anchors[anchor_step * n + 1]
        anchor_box = [0, 0, aw, ah]
        iou = bbox_iou(anchor_box, gt_box, x1y1x2y2=False)
        if anchor_step == 4:
          ax = anchors[anchor_step * n + 2]
          ay = anchors[anchor_step * n + 3]
          dist = pow(((gi + ax) - gx), 2) + pow(((gj + ay) - gy), 2)
        if iou > best_iou:
          best_iou = iou
          best_n = n
        elif anchor_step == 4 and iou == best_iou and dist < min_dist:
          best_iou = iou
          best_n = n
          min_dist = dist

      gt_box = [gx, gy, gw, gh]
      pred_box = pred_boxes[b * nAnchors + best_n * nPixels + gj * nW + gi]

      coord_mask[b][best_n][gj][gi] = 1
      cls_mask[b][best_n][gj][gi] = 1
      conf_mask[b][best_n][gj][gi] = object_scale
      tx[b][best_n][gj][gi] = target[b][t * 5 + 1] * nW - gi
      ty[b][best_n][gj][gi] = target[b][t * 5 + 2] * nH - gj
      tw[b][best_n][gj][gi] = math.log(gw / anchors[anchor_step * best_n])
      th[b][best_n][gj][gi] = math.log(gh / anchors[anchor_step * best_n + 1])
      iou = bbox_iou(gt_box, pred_box, x1y1x2y2=False)  # best_iou
      tconf[b][best_n][gj][gi] = iou
      tcls[b][best_n][gj][gi] = target[b][t * 5]
      if iou > 0.5:
        nCorrect = nCorrect + 1

  return nGT, nCorrect, coord_mask, conf_mask, cls_mask, tx, ty, tw, th, tconf, tcls


class RegionLoss(nn.Module):
  def __init__(self, num_classes=0, anchors=[], num_anchors=1):
    super(RegionLoss, self).__init__()
    self.num_classes = num_classes
    self.anchors = anchors
    self.num_anchors = num_anchors
    self.anchor_step = len(anchors) / num_anchors
    self.coord_scale = 1
    self.noobject_scale = 1
    self.object_scale = 5
    self.class_scale = 1
    self.thresh = 0.6
    self.seen = 0

  def forward(self, output, target):
    # output : BxAs*(4+1+num_classes)*H*W
    t0 = time.time()
    nB = output.data.size(0)
    nA = self.num_anchors
    nC = self.num_classes
    nH = output.data.size(2)
    nW = output.data.size(3)

    output = output.view(nB, nA, (5 + nC), nH, nW)
    x = F.sigmoid(output.index_select(2, Variable(torch.cuda.LongTensor([0]))).view(nB, nA, nH, nW))
    y = F.sigmoid(output.index_select(2, Variable(torch.cuda.LongTensor([1]))).view(nB, nA, nH, nW))
    w = output.index_select(2, Variable(torch.cuda.LongTensor([2]))).view(nB, nA, nH, nW)
    h = output.index_select(2, Variable(torch.cuda.LongTensor([3]))).view(nB, nA, nH, nW)
    conf = F.sigmoid(output.index_select(2, Variable(torch.cuda.LongTensor([4]))).view(nB, nA, nH, nW))
    cls = output.index_select(2, Variable(torch.linspace(5, 5 + nC - 1, nC).long().cuda()))
    cls = cls.view(nB * nA, nC, nH * nW).transpose(1, 2).contiguous().view(nB * nA * nH * nW, nC)
    t1 = time.time()

    pred_boxes = torch.cuda.FloatTensor(4, nB * nA * nH * nW)
    grid_x = torch.linspace(0, nW - 1, nW).repeat(nH, 1).repeat(nB * nA, 1, 1).view(nB * nA * nH * nW).cuda()
    grid_y = torch.linspace(0, nH - 1, nH).repeat(nW, 1).t().repeat(nB * nA, 1, 1).view(nB * nA * nH * nW).cuda()
    anchor_w = torch.Tensor(self.anchors).view(nA, self.anchor_step).index_select(1, torch.LongTensor([0])).cuda()
    anchor_h = torch.Tensor(self.anchors).view(nA, self.anchor_step).index_select(1, torch.LongTensor([1])).cuda()
    anchor_w = anchor_w.repeat(nB, 1).repeat(1, 1, nH * nW).view(nB * nA * nH * nW)
    anchor_h = anchor_h.repeat(nB, 1).repeat(1, 1, nH * nW).view(nB * nA * nH * nW)
    pred_boxes[0] = x.data + grid_x
    pred_boxes[1] = y.data + grid_y
    pred_boxes[2] = torch.exp(w.data) * anchor_w
    pred_boxes[3] = torch.exp(h.data) * anchor_h
    pred_boxes = convert2cpu(pred_boxes.transpose(0, 1).contiguous().view(-1, 4))
    t2 = time.time()

    nGT, nCorrect, coord_mask, conf_mask, cls_mask, tx, ty, tw, th, tconf, tcls = build_targets(pred_boxes,
                                                  target.data,
                                                  self.anchors, nA,
                                                  nC, \
                                                  nH, nW,
                                                  self.noobject_scale,
                                                  self.object_scale,
                                                  self.thresh,
                                                  self.seen)
    cls_mask = (cls_mask == 1)
    nProposals = int((conf > 0.25).sum().data[0])

    tx = Variable(tx.cuda())
    ty = Variable(ty.cuda())
    tw = Variable(tw.cuda())
    th = Variable(th.cuda())
    tconf = Variable(tconf.cuda())
    tcls = Variable(tcls.view(-1)[cls_mask].long().cuda())

    coord_mask = Variable(coord_mask.cuda())
    conf_mask = Variable(conf_mask.cuda().sqrt())
    cls_mask = Variable(cls_mask.view(-1, 1).repeat(1, nC).cuda())
    cls = cls[cls_mask].view(-1, nC)

    t3 = time.time()

    loss_x = self.coord_scale * nn.MSELoss(reduction='sum')(x * coord_mask, tx * coord_mask) / 2.0
    loss_y = self.coord_scale * nn.MSELoss(reduction='sum')(y * coord_mask, ty * coord_mask) / 2.0
    loss_w = self.coord_scale * nn.MSELoss(reduction='sum')(w * coord_mask, tw * coord_mask) / 2.0
    loss_h = self.coord_scale * nn.MSELoss(reduction='sum')(h * coord_mask, th * coord_mask) / 2.0
    loss_conf = nn.MSELoss(reduction='sum')(conf * conf_mask, tconf * conf_mask) / 2.0
    loss_cls = self.class_scale * nn.CrossEntropyLoss(reduction='sum')(cls, tcls)
    loss = loss_x + loss_y + loss_w + loss_h + loss_conf + loss_cls
    t4 = time.time()
    if False:
      print('-----------------------------------')
      print('    activation : %f' % (t1 - t0))
      print(' create pred_boxes : %f' % (t2 - t1))
      print('   build targets : %f' % (t3 - t2))
      print('     create loss : %f' % (t4 - t3))
      print('       total : %f' % (t4 - t0))
    print('%d: nGT %d, recall %d, proposals %d, loss: x %f, y %f, w %f, h %f, conf %f, cls %f, total %f' % (
    self.seen, nGT, nCorrect, nProposals, loss_x.data[0], loss_y.data[0], loss_w.data[0], loss_h.data[0],
    loss_conf.data[0], loss_cls.data[0], loss.data[0]))
    return loss



import torch.nn as nn
import torch.nn.functional as F
from darknet2any.tool.torch_utils import *

def yolo_forward(output, conf_thresh, num_classes, anchors, num_anchors, scale_x_y, only_objectness=1,
                validation=False):
  # Output would be invalid if it does not satisfy this assert
  # assert (output.size(1) == (5 + num_classes) * num_anchors)

  # print(output.size())

  # Slice the second dimension (channel) of output into:
  # [ 2, 2, 1, num_classes, 2, 2, 1, num_classes, 2, 2, 1, num_classes ]
  # And then into
  # bxy = [ 6 ] bwh = [ 6 ] det_conf = [ 3 ] cls_conf = [ num_classes * 3 ]
  batch = output.size(0)
  H = output.size(2)
  W = output.size(3)

  bxy_list = []
  bwh_list = []
  det_confs_list = []
  cls_confs_list = []

  for i in range(num_anchors):
    begin = i * (5 + num_classes)
    end = (i + 1) * (5 + num_classes)
    
    bxy_list.append(output[:, begin : begin + 2])
    bwh_list.append(output[:, begin + 2 : begin + 4])
    det_confs_list.append(output[:, begin + 4 : begin + 5])
    cls_confs_list.append(output[:, begin + 5 : end])

  # Shape: [batch, num_anchors * 2, H, W]
  bxy = torch.cat(bxy_list, dim=1)
  # Shape: [batch, num_anchors * 2, H, W]
  bwh = torch.cat(bwh_list, dim=1)

  # Shape: [batch, num_anchors, H, W]
  det_confs = torch.cat(det_confs_list, dim=1)
  # Shape: [batch, num_anchors * H * W]
  det_confs = det_confs.view(batch, num_anchors * H * W)

  # Shape: [batch, num_anchors * num_classes, H, W]
  cls_confs = torch.cat(cls_confs_list, dim=1)
  # Shape: [batch, num_anchors, num_classes, H * W]
  cls_confs = cls_confs.view(batch, num_anchors, num_classes, H * W)
  # Shape: [batch, num_anchors, num_classes, H * W] --> [batch, num_anchors * H * W, num_classes] 
  cls_confs = cls_confs.permute(0, 1, 3, 2).reshape(batch, num_anchors * H * W, num_classes)

  # Apply sigmoid(), exp() and softmax() to slices
  #
  bxy = torch.sigmoid(bxy) * scale_x_y - 0.5 * (scale_x_y - 1)
  bwh = torch.exp(bwh)
  det_confs = torch.sigmoid(det_confs)
  cls_confs = torch.sigmoid(cls_confs)

  # Prepare C-x, C-y, P-w, P-h (None of them are torch related)
  grid_x = np.expand_dims(np.expand_dims(np.expand_dims(np.linspace(0, W - 1, W), axis=0).repeat(H, 0), axis=0), axis=0)
  grid_y = np.expand_dims(np.expand_dims(np.expand_dims(np.linspace(0, H - 1, H), axis=1).repeat(W, 1), axis=0), axis=0)
  # grid_x = torch.linspace(0, W - 1, W).reshape(1, 1, 1, W).repeat(1, 1, H, 1)
  # grid_y = torch.linspace(0, H - 1, H).reshape(1, 1, H, 1).repeat(1, 1, 1, W)

  anchor_w = []
  anchor_h = []
  for i in range(num_anchors):
    anchor_w.append(anchors[i * 2])
    anchor_h.append(anchors[i * 2 + 1])

  device = None
  cuda_check = output.is_cuda
  if cuda_check:
    device = output.get_device()

  bx_list = []
  by_list = []
  bw_list = []
  bh_list = []

  # Apply C-x, C-y, P-w, P-h
  for i in range(num_anchors):
    ii = i * 2
    # Shape: [batch, 1, H, W]
    bx = bxy[:, ii : ii + 1] + torch.tensor(grid_x, device=device, dtype=torch.float32) # grid_x.to(device=device, dtype=torch.float32)
    # Shape: [batch, 1, H, W]
    by = bxy[:, ii + 1 : ii + 2] + torch.tensor(grid_y, device=device, dtype=torch.float32) # grid_y.to(device=device, dtype=torch.float32)
    # Shape: [batch, 1, H, W]
    bw = bwh[:, ii : ii + 1] * anchor_w[i]
    # Shape: [batch, 1, H, W]
    bh = bwh[:, ii + 1 : ii + 2] * anchor_h[i]

    bx_list.append(bx)
    by_list.append(by)
    bw_list.append(bw)
    bh_list.append(bh)


  ########################################
  #   Figure out bboxes from slices   #
  ########################################
  
  # Shape: [batch, num_anchors, H, W]
  bx = torch.cat(bx_list, dim=1)
  # Shape: [batch, num_anchors, H, W]
  by = torch.cat(by_list, dim=1)
  # Shape: [batch, num_anchors, H, W]
  bw = torch.cat(bw_list, dim=1)
  # Shape: [batch, num_anchors, H, W]
  bh = torch.cat(bh_list, dim=1)

  # Shape: [batch, 2 * num_anchors, H, W]
  bx_bw = torch.cat((bx, bw), dim=1)
  # Shape: [batch, 2 * num_anchors, H, W]
  by_bh = torch.cat((by, bh), dim=1)

  # normalize coordinates to [0, 1]
  bx_bw /= W
  by_bh /= H

  # Shape: [batch, num_anchors * H * W, 1]
  bx = bx_bw[:, :num_anchors].view(batch, num_anchors * H * W, 1)
  by = by_bh[:, :num_anchors].view(batch, num_anchors * H * W, 1)
  bw = bx_bw[:, num_anchors:].view(batch, num_anchors * H * W, 1)
  bh = by_bh[:, num_anchors:].view(batch, num_anchors * H * W, 1)

  bx1 = bx - bw * 0.5
  by1 = by - bh * 0.5
  bx2 = bx1 + bw
  by2 = by1 + bh

  # Shape: [batch, num_anchors * h * w, 4] -> [batch, num_anchors * h * w, 1, 4]
  boxes = torch.cat((bx1, by1, bx2, by2), dim=2).view(batch, num_anchors * H * W, 1, 4)
  # boxes = boxes.repeat(1, 1, num_classes, 1)

  # boxes:   [batch, num_anchors * H * W, 1, 4]
  # cls_confs: [batch, num_anchors * H * W, num_classes]
  # det_confs: [batch, num_anchors * H * W]

  det_confs = det_confs.view(batch, num_anchors * H * W, 1)
  confs = cls_confs * det_confs

  # boxes: [batch, num_anchors * H * W, 1, 4]
  # confs: [batch, num_anchors * H * W, num_classes]

  return  boxes, confs, output


def yolo_forward_dynamic(output, conf_thresh, num_classes, anchors, num_anchors, scale_x_y, only_objectness=1,
                validation=False):
  # Output would be invalid if it does not satisfy this assert
  # assert (output.size(1) == (5 + num_classes) * num_anchors)

  # print(output.size())

  # Slice the second dimension (channel) of output into:
  # [ 2, 2, 1, num_classes, 2, 2, 1, num_classes, 2, 2, 1, num_classes ]
  # And then into
  # bxy = [ 6 ] bwh = [ 6 ] det_conf = [ 3 ] cls_conf = [ num_classes * 3 ]
  # batch = output.size(0)
  # H = output.size(2)
  # W = output.size(3)

  bxy_list = []
  bwh_list = []
  det_confs_list = []
  cls_confs_list = []

  for i in range(num_anchors):
    begin = i * (5 + num_classes)
    end = (i + 1) * (5 + num_classes)
    
    bxy_list.append(output[:, begin : begin + 2])
    bwh_list.append(output[:, begin + 2 : begin + 4])
    det_confs_list.append(output[:, begin + 4 : begin + 5])
    cls_confs_list.append(output[:, begin + 5 : end])

  # Shape: [batch, num_anchors * 2, H, W]
  bxy = torch.cat(bxy_list, dim=1)
  # Shape: [batch, num_anchors * 2, H, W]
  bwh = torch.cat(bwh_list, dim=1)

  # Shape: [batch, num_anchors, H, W]
  det_confs = torch.cat(det_confs_list, dim=1)
  # Shape: [batch, num_anchors * H * W]
  det_confs = det_confs.view(output.size(0), num_anchors * output.size(2) * output.size(3))

  # Shape: [batch, num_anchors * num_classes, H, W]
  cls_confs = torch.cat(cls_confs_list, dim=1)
  # Shape: [batch, num_anchors, num_classes, H * W]
  cls_confs = cls_confs.view(
    output.size(0), num_anchors, num_classes, output.size(2) * output.size(3))
  # Shape: [batch, num_anchors, num_classes, H * W] --> [batch, num_anchors * H * W, num_classes] 
  cls_confs = cls_confs.permute(0, 1, 3, 2).reshape(
    output.size(0), num_anchors * output.size(2) * output.size(3), num_classes)

  # Apply sigmoid(), exp() and softmax() to slices
  #
  bxy = torch.sigmoid(bxy) * scale_x_y - 0.5 * (scale_x_y - 1)
  bwh = torch.exp(bwh)
  det_confs = torch.sigmoid(det_confs)
  cls_confs = torch.sigmoid(cls_confs)

  # Prepare C-x, C-y, P-w, P-h (None of them are torch related)
  grid_x = np.expand_dims(
    np.expand_dims(
      np.expand_dims(
        np.linspace(0, output.size(3) - 1, output.size(3)), axis=0).repeat(
          output.size(2), 0), axis=0), axis=0)
  grid_y = np.expand_dims(
    np.expand_dims(
      np.expand_dims(
        np.linspace(0, output.size(2) - 1, output.size(2)), axis=1).repeat(
          output.size(3), 1), axis=0), axis=0)
  # grid_x = torch.linspace(0, W - 1, W).reshape(1, 1, 1, W).repeat(1, 1, H, 1)
  # grid_y = torch.linspace(0, H - 1, H).reshape(1, 1, H, 1).repeat(1, 1, 1, W)

  anchor_w = []
  anchor_h = []
  for i in range(num_anchors):
    anchor_w.append(anchors[i * 2])
    anchor_h.append(anchors[i * 2 + 1])

  device = None
  cuda_check = output.is_cuda
  if cuda_check:
    device = output.get_device()

  bx_list = []
  by_list = []
  bw_list = []
  bh_list = []

  # Apply C-x, C-y, P-w, P-h
  for i in range(num_anchors):
    ii = i * 2
    # Shape: [batch, 1, H, W]
    bx = bxy[:, ii : ii + 1] + torch.tensor(grid_x, device=device, dtype=torch.float32)
    # Shape: [batch, 1, H, W]
    by = bxy[:, ii + 1 : ii + 2] + torch.tensor(grid_y, device=device, dtype=torch.float32)
    # Shape: [batch, 1, H, W]
    bw = bwh[:, ii : ii + 1] * anchor_w[i]
    # Shape: [batch, 1, H, W]
    bh = bwh[:, ii + 1 : ii + 2] * anchor_h[i]

    bx_list.append(bx)
    by_list.append(by)
    bw_list.append(bw)
    bh_list.append(bh)


  ########################################
  #   Figure out bboxes from slices   #
  ########################################
  
  # Shape: [batch, num_anchors, H, W]
  bx = torch.cat(bx_list, dim=1)
  # Shape: [batch, num_anchors, H, W]
  by = torch.cat(by_list, dim=1)
  # Shape: [batch, num_anchors, H, W]
  bw = torch.cat(bw_list, dim=1)
  # Shape: [batch, num_anchors, H, W]
  bh = torch.cat(bh_list, dim=1)

  # Shape: [batch, 2 * num_anchors, H, W]
  bx_bw = torch.cat((bx, bw), dim=1)
  # Shape: [batch, 2 * num_anchors, H, W]
  by_bh = torch.cat((by, bh), dim=1)

  # normalize coordinates to [0, 1]
  bx_bw /= output.size(3)
  by_bh /= output.size(2)

  # Shape: [batch, num_anchors * H * W, 1]
  bx = bx_bw[:, :num_anchors].view(
    output.size(0), num_anchors * output.size(2) * output.size(3), 1)
  by = by_bh[:, :num_anchors].view(
    output.size(0), num_anchors * output.size(2) * output.size(3), 1)
  bw = bx_bw[:, num_anchors:].view(
    output.size(0), num_anchors * output.size(2) * output.size(3), 1)
  bh = by_bh[:, num_anchors:].view(
    output.size(0), num_anchors * output.size(2) * output.size(3), 1)

  bx1 = bx - bw * 0.5
  by1 = by - bh * 0.5
  bx2 = bx1 + bw
  by2 = by1 + bh

  # Shape: [batch, num_anchors * h * w, 4] -> [batch, num_anchors * h * w, 1, 4]
  boxes = torch.cat((bx1, by1, bx2, by2), dim=2).view(
    output.size(0), num_anchors * output.size(2) * output.size(3), 1, 4)
  # boxes = boxes.repeat(1, 1, num_classes, 1)

  # boxes:   [batch, num_anchors * H * W, 1, 4]
  # cls_confs: [batch, num_anchors * H * W, num_classes]
  # det_confs: [batch, num_anchors * H * W]

  det_confs = det_confs.view(
    output.size(0), num_anchors * output.size(2) * output.size(3), 1)
  confs = cls_confs * det_confs

  # boxes: [batch, num_anchors * H * W, 1, 4]
  # confs: [batch, num_anchors * H * W, num_classes]

  return  boxes, confs, output

class YoloLayer(nn.Module):
  ''' Yolo layer
  model_out: while inference,is post-processing inside or outside the model
    true:outside
  '''
  def __init__(self, anchor_mask=[], num_classes=0, anchors=[],
    num_anchors=1, stride=32, model_out=False):
    
    super(YoloLayer, self).__init__()
    self.anchor_mask = anchor_mask
    self.num_classes = num_classes
    self.anchors = anchors
    self.num_anchors = num_anchors
    self.anchor_step = len(anchors) // num_anchors
    self.coord_scale = 1
    self.noobject_scale = 1
    self.object_scale = 5
    self.class_scale = 1
    self.thresh = 0.6
    self.stride = stride
    self.seen = 0
    self.scale_x_y = 1

    self.model_out = model_out

  def forward(self, output, target=None):
    if self.training:
      return output
    masked_anchors = []
    for m in self.anchor_mask:
      masked_anchors += self.anchors[m * self.anchor_step:(m + 1) * self.anchor_step]
    masked_anchors = [anchor / self.stride for anchor in masked_anchors]

    return yolo_forward_dynamic(output, self.thresh, self.num_classes,
      masked_anchors, len(self.anchor_mask),scale_x_y=self.scale_x_y)




import re
import sys
import os
import time
import math
import numpy as np

import itertools
import struct  # get_image_size
import imghdr  # get_image_size

ENDS_WITH_NUMBER = re.compile(r'(.*)_[0-9]+')

def get_darknet_files(weights_file):
  """
  return an appropriate weights, cfg, and names file if they exist.
  Returns:
    (str, str, str, str): weights, cfg, names, prefix for other files
  """
  
  if not weights_file.endswith(".weights"):
    weights_file += ".weights"

  prefix = os.path.splitext(weights_file)[0]
  original = prefix
  
  cfg_file = f"{prefix}.cfg"
  names_file = f"{prefix}.names"

  if os.path.isfile(weights_file):
    if not os.path.isfile(cfg_file) or not os.path.isfile(names_file):
      
      if prefix.endswith("_best"):
        prefix = prefix.replace("_best", "")

      elif prefix.endswith("_last"):
        prefix = prefix.replace("_last", "")

      elif prefix.endswith("_final"):
        prefix = prefix.replace("_final", "")

      else:
        m = ENDS_WITH_NUMBER.match(prefix)
        
        if m:
          prefix = m.group(1)
      
      if original != prefix:
        cfg_file = f"{prefix}.cfg"
        names_file = f"{prefix}.names"
      
  if not (os.path.isfile(cfg_file) and os.path.isfile(cfg_file) and
    os.path.isfile(weights_file)):
  
    cfg_file = None
    names_file = None
    weights_file = None
  
  return weights_file, cfg_file, names_file, prefix

def sigmoid(x):
  return 1.0 / (np.exp(-x) + 1.)


def softmax(x):
  x = np.exp(x - np.expand_dims(np.max(x, axis=1), axis=1))
  x = x / np.expand_dims(x.sum(axis=1), axis=1)
  return x


def bbox_iou(box1, box2, x1y1x2y2=True):
  
  # print('iou box1:', box1)
  # print('iou box2:', box2)

  if x1y1x2y2:
    mx = min(box1[0], box2[0])
    Mx = max(box1[2], box2[2])
    my = min(box1[1], box2[1])
    My = max(box1[3], box2[3])
    w1 = box1[2] - box1[0]
    h1 = box1[3] - box1[1]
    w2 = box2[2] - box2[0]
    h2 = box2[3] - box2[1]
  else:
    w1 = box1[2]
    h1 = box1[3]
    w2 = box2[2]
    h2 = box2[3]

    mx = min(box1[0], box2[0])
    Mx = max(box1[0] + w1, box2[0] + w2)
    my = min(box1[1], box2[1])
    My = max(box1[1] + h1, box2[1] + h2)
  uw = Mx - mx
  uh = My - my
  cw = w1 + w2 - uw
  ch = h1 + h2 - uh
  carea = 0
  if cw <= 0 or ch <= 0:
    return 0.0

  area1 = w1 * h1
  area2 = w2 * h2
  carea = cw * ch
  uarea = area1 + area2 - carea
  return carea / uarea


def nms_cpu(boxes, confs, nms_thresh=0.5, min_mode=False):
  # print(boxes.shape)
  x1 = boxes[:, 0]
  y1 = boxes[:, 1]
  x2 = boxes[:, 2]
  y2 = boxes[:, 3]

  areas = (x2 - x1) * (y2 - y1)
  order = confs.argsort()[::-1]

  keep = []
  while order.size > 0:
    idx_self = order[0]
    idx_other = order[1:]

    keep.append(idx_self)

    xx1 = np.maximum(x1[idx_self], x1[idx_other])
    yy1 = np.maximum(y1[idx_self], y1[idx_other])
    xx2 = np.minimum(x2[idx_self], x2[idx_other])
    yy2 = np.minimum(y2[idx_self], y2[idx_other])

    w = np.maximum(0.0, xx2 - xx1)
    h = np.maximum(0.0, yy2 - yy1)
    inter = w * h

    if min_mode:
      over = inter / np.minimum(areas[order[0]], areas[order[1:]])
    else:
      over = inter / (areas[order[0]] + areas[order[1:]] - inter)

    inds = np.where(over <= nms_thresh)[0]
    order = order[inds + 1]
  
  return np.array(keep)



def plot_boxes_cv2(img, boxes, savename=None, class_names=None, color=None):
  import cv2
  img = np.copy(img)
  colors = np.array([[1, 0, 1], [0, 0, 1], [0, 1, 1], [0, 1, 0], [1, 1, 0], [1, 0, 0]], dtype=np.float32)

  def get_color(c, x, max_val):
    ratio = float(x) / max_val * 5
    i = int(math.floor(ratio))
    j = int(math.ceil(ratio))
    ratio = ratio - i
    r = (1 - ratio) * colors[i][c] + ratio * colors[j][c]
    return int(r * 255)

  width = img.shape[1]
  height = img.shape[0]
  for i in range(len(boxes)):
    box = boxes[i]
    x1 = int(box[0] * width)
    y1 = int(box[1] * height)
    x2 = int(box[2] * width)
    y2 = int(box[3] * height)
    bbox_thick = int(0.6 * (height + width) / 600)
    if color:
      rgb = color
    else:
      rgb = (255, 0, 0)
    if len(box) >= 7 and class_names:
      cls_conf = box[5]
      cls_id = box[6]
      print(f"{class_names[cls_id]} ({cls_conf}) at coords: {x1},{y1}->{x2},{y2}")
      classes = len(class_names)
      offset = cls_id * 123457 % classes
      red = get_color(2, offset, classes)
      green = get_color(1, offset, classes)
      blue = get_color(0, offset, classes)
      if color is None:
        rgb = (red, green, blue)
      msg = str(class_names[cls_id])+" "+str(round(cls_conf,3))
      t_size = cv2.getTextSize(msg, 0, 0.7, thickness=bbox_thick // 2)[0]
      c1, c2 = (x1,y1), (x2, y2)
      c3 = (c1[0] + t_size[0], c1[1] - t_size[1] - 3)
      print(f"c1={c1}, c2={c2}, c3={c3}, rgb={rgb}, ")
      cv2.rectangle(img, (x1,y1), (c3[0], c3[1]), rgb, -1)
      #cv2.rectangle(img, (x1,y1), (np.float32(c3[0]), np.float32(c3[1])), rgb, -1)
      img = cv2.putText(img, msg, (c1[0], c1[1] - 2), cv2.FONT_HERSHEY_SIMPLEX,0.7, (0,0,0), bbox_thick//2,lineType=cv2.LINE_AA)
    
    img = cv2.rectangle(img, (x1, y1), (x2, y2), rgb, bbox_thick)
  if savename:
    #print("save plot results to %s" % savename)
    cv2.imwrite(savename, img)
  return img


def read_truths(lab_path):
  if not os.path.exists(lab_path):
    return np.array([])
  if os.path.getsize(lab_path):
    truths = np.loadtxt(lab_path)
    truths = truths.reshape(truths.size / 5, 5)  # to avoid single truth problem
    return truths
  else:
    return np.array([])


def load_class_names(namesfile):
  class_names = []
  with open(namesfile, 'r') as fp:
    lines = fp.readlines()
  for line in lines:
    line = line.rstrip()
    class_names.append(line)
  return class_names



def post_processing(img, conf_thresh, nms_thresh, output, debug=False):

  # anchors = [12, 16, 19, 36, 40, 28, 36, 75, 76, 55, 72, 146, 142, 110, 192, 243, 459, 401]
  # num_anchors = 9
  # anchor_masks = [[0, 1, 2], [3, 4, 5], [6, 7, 8]]
  # strides = [8, 16, 32]
  # anchor_step = len(anchors) // num_anchors

  # [batch, num, 1, 4]
  box_array = output[0]
  # [batch, num, num_classes]
  confs = output[1]

  t1 = time.time()

  if type(box_array).__name__ != 'ndarray':
    box_array = box_array.cpu().detach().numpy()
    confs = confs.cpu().detach().numpy()

  num_classes = confs.shape[2]

  if debug:
    print(f"post_processing: using num_classes={num_classes}")

  # [batch, num, 4]
  box_array = box_array[:, :, 0]

  # [batch, num, num_classes] --> [batch, num]
  max_conf = np.max(confs, axis=2)
  max_id = np.argmax(confs, axis=2)

  t2 = time.time()

  bboxes_batch = []
  for i in range(box_array.shape[0]):
     
    argwhere = max_conf[i] > conf_thresh
    l_box_array = box_array[i, argwhere, :]
    l_max_conf = max_conf[i, argwhere]
    l_max_id = max_id[i, argwhere]

    bboxes = []
    # nms for each class
    for j in range(num_classes):

      cls_argwhere = l_max_id == j
      ll_box_array = l_box_array[cls_argwhere, :]
      ll_max_conf = l_max_conf[cls_argwhere]
      ll_max_id = l_max_id[cls_argwhere]

      keep = nms_cpu(ll_box_array, ll_max_conf, nms_thresh)
      
      if (keep.size > 0):
        ll_box_array = ll_box_array[keep, :]
        ll_max_conf = ll_max_conf[keep]
        ll_max_id = ll_max_id[keep]

        for k in range(ll_box_array.shape[0]):
          bboxes.append([ll_box_array[k, 0], ll_box_array[k, 1], ll_box_array[k, 2], ll_box_array[k, 3], ll_max_conf[k], ll_max_conf[k], ll_max_id[k]])
    
    bboxes_batch.append(bboxes)

  t3 = time.time()

  if debug:
    print('-----------------------------------')
    print('       max and argmax : %f' % (t2 - t1))
    print('                  nms : %f' % (t3 - t2))
    print('Post processing total : %f' % (t3 - t1))
    print('-----------------------------------')
  
  return bboxes_batch



############################################################################
# --------------------------------------------------------------------------
# @author James Edmondson <james@koshee.ai>
# --------------------------------------------------------------------------
############################################################################

"""
provides conversion from darknet to onnx format
"""

import sys
import onnx
import os
import argparse
import numpy as np
import cv2
import onnxruntime

from darknet2any.tool.utils import *
from darknet2any.tool.darknet2onnx import *

def parse_args(args):
  """
  parses command line arguments

  Args:
  args (list): list of arguments
  Returns:
  dict: a map of arguments as defined by the parser
  """
  parser = argparse.ArgumentParser(
  description="Converts a yolov4 weights file to onnx",
  add_help=True
  )
  parser.add_argument('-i','--input','--weights', action='store',
    dest='input', default=None,
    help='the weights file to convert')
  parser.add_argument('--image','--test', action='store',
    dest='image', default=None,
    help='the image to test the resulting onnx model on')
  parser.add_argument('--ie','--image-embeddings','--include-embeddings',
    action='store_true', dest='include_embeddings',
    help='add image feature embeddings to model output')
  parser.add_argument('-o','--output','--onnx',
    action='store', dest='output', default=None,
    help='the onnx file to create (default=filename.onnx)')
  parser.add_argument('--op','--opset',
    action='store', type=int, dest='opset', default=15,
    help='the opset to use for export. 15 is default.')

  return parser.parse_args(args)

def convert(cfg_file, weight_file, output_name, include_embeddings, opset):
  """
  converts the darknet model
  """

  transform_to_onnx(cfg_file, weight_file, 1, output_name, include_embeddings, opset)

def main():
  """
  main script entry point
  """
  options = parse_args(sys.argv[1:])

  if options.input is not None:
    prefix = os.path.splitext(options.input)[0]

    if not os.path.isfile(options.input):
      print(f"darknet2onnx: darknet weights file cannot be read. "
        "check file exists or permissions.")
      exit(1)

    weights_file, cfg_file, names_file, prefix = get_darknet_files(options.input)

    image_path = options.image
    batch_size = 1
    output_file = f"{prefix}.onnx"

    if weights_file is not None:
      print(f"darknet2onnx: converting darknet weights to onnx...")
      print(f"  weights_file={weights_file}")
      print(f"  names_file={names_file}")
      print(f"  cfg_file={cfg_file}")
      print(f"  target={output_file}")
      print(f"  opset={options.opset}")

      if options.output is not None:
        output_file = options.output

      convert(cfg_file, weights_file, output_file, options.include_embeddings, options.opset)

      print("darknet2onnx: conversion complete")
    else:
      print("darknet2onnx: unable to find appropriate names/cfg files for "
        "weights file. Ideally, name.weights should have name.cfg and name.names"
        "in the same directory")
  else:
    parse_args(["-h"])


if __name__ == '__main__':
  main()



############################################################################
# --------------------------------------------------------------------------
# @author James Edmondson <james@koshee.ai>
# --------------------------------------------------------------------------
############################################################################

"""
predicts using onnx on a directory structure of images
"""

import argparse
import importlib
import numpy as np
import os
import sys
import time
from threading import Thread

import cv2

from darknet2any.tool.utils import *
from darknet2any.tool.darknet2onnx import *

ort_loader = importlib.util.find_spec('onnxruntime')

if not ort_loader:
  print(f"darknet2any: this script requires an installation with onnxruntime")
  print(f"  try pip install darknet2any[tensorrt] or [rocm] options")

  exit(0)

import onnxruntime as ort

available_providers = ort.get_available_providers()

print(f"onnx executor options: {available_providers}")

class Stats:
  """
  a stats class for aggregating thread metrics
  """
  def __init__(self):
    """
    constructor
    """
    self.embed_time = 0
    self.fps = 0
    self.load_time = 0
    self.read_time = 0
    self.predict_time = 0
    self.process_time = 0
    self.write_time = 0
    self.context_create_time = 0
    self.avg_embed_time = 0
    self.avg_load_time = 0
    self.avg_read_time = 0
    self.avg_predict_time = 0
    self.avg_process_time = 0
    self.avg_write_time = 0
    self.avg_context_create_time = 0
    self.num_predicts = 0
    self.num_threads = 0
  
  def add(self, thread):
    """
    adds thread stats to the aggregated stats
    """
    self.embed_time += thread.embed_time
    self.load_time += thread.load_time
    self.num_predicts += thread.num_predicts
    self.read_time += thread.read_time
    self.predict_time += thread.predict_time
    self.process_time += thread.process_time
    self.write_time += thread.write_time
    self.context_create_time += thread.context_create_time
    self.num_threads += 1

  def print(self, total_time):
    """
    prints the aggregated stats
    """
    self.fps = self.num_predicts / total_time
    self.avg_embed_time = self.embed_time / self.num_predicts
    self.avg_load_time = self.load_time / self.num_predicts
    self.avg_read_time = self.read_time / self.num_predicts
    self.avg_predict_time = self.predict_time / self.num_predicts
    self.avg_process_time = self.process_time / self.num_predicts
    self.avg_write_time = self.write_time / self.num_predicts
    self.avg_context_create_time = self.context_create_time / self.num_threads
    
    print("Overall stats:")
    print(f"  predicts: {self.num_predicts}, time: {total_time:.3f}s, "
      f"fps: {self.fps:.2f}")
    print(f"  load_time: time: {self.load_time:.3f}s, avg: {self.avg_load_time:.4f}s")
    print(f"  reads: time: {self.read_time:.3f}s, avg: {self.avg_read_time:.4f}s")
    print(f"  predicts: time: {self.predict_time:.4f}s, "
      f"avg: {self.avg_predict_time:.4f}s")
    print(f"  writes: time: {self.write_time:.4f}s, "
      f"avg: {self.avg_write_time:.4f}s")
    print(f"  embed: time: {self.embed_time:.4f}s, "
      f"avg: {self.avg_embed_time:.4f}s")

class OnnxThread(Thread):
  """
  threaded onnxruntime implementation
  """
  def __init__(self, idx, model_path, providers, classes,
    output, num_threads, images):
    """
    constructor
    """
    
    Thread.__init__(self)
    
    self.avg_embed_time = 0
    self.avg_predict_time = 0
    self.avg_process_time = 0
    self.avg_read_time = 0
    self.avg_write_time = 0
    self.bindings = list()
    self.buffers = None
    self.classes = classes
    self.context_create_time = 0
    self.embed_time = 0
    self.model_path = model_path
    self.device = None
    self.session = None
    self.fps = 0
    self.idx = idx
    self.images = self.get_images(images, num_threads)
    self.inputs = list()
    self.load_time = 0
    self.num_predicts = 0
    self.num_threads = num_threads
    self.output_dir = output
    self.outputs = list()
    self.predict_time = 0
    self.providers = providers
    self.process_time = 0
    self.read_time = 0
    self.shape = None
    self.stream = None
    self.trt_context = None
    self.write_time = 0
    self.in_run = False

  def get_images(self, images, num_threads):
    '''
    Allocates an images list based on num threads and length of images
    '''
    segment = int(len(images) / num_threads)
    start = int(segment * self.idx)
    end = int(segment) * (self.idx + 1)
    
    results = None
    
    if self.idx != num_threads - 1:
      results = images[start:end]
    else:
      results = images[start:]
    
    #print(f"get_images: images={results}")
    
    return results
  
  def load(self):
    
      print(f"predict_onnx.{self.idx}: using providers={self.providers}")
      print(f"predict_onnx.{self.idx}: loading model={self.model_path}")

      # 1. Load the onnx model
      start = time.perf_counter()
      self.session = ort.InferenceSession(self.model_path,
        providers=self.providers)
      end = time.perf_counter()
      self.load_time = end - start
      print(f"  load_time: {self.load_time:.4f}s")
      
      print(f"  provider options: {self.session.get_provider_options()}")
      
      self.shape = None

      for input_meta in self.session.get_inputs():
        if input_meta.name == "input":
          self.shape = (
            input_meta.shape[3],
            input_meta.shape[2]
          )
          break


  def predict(self, image_file):
    """
    predicts classes of an image file

    Args:
    interpreter (tf.lite.Interpreter): the onnx interpreter for a model
    input_details (list[dict[string, Any]]): result of get_input_details()
    output_details (list[dict[string, Any]]): result of get_output_details()
    image_file (str): an image file to read and predict on
    Returns:
    tuple: read_time, predict_time
    """
    print(f"onnx: Reading {image_file}")
    self.num_predicts += 1
    basename = os.path.basename(image_file)

    start = time.perf_counter()
    img = cv2.imread(image_file)
    image_rgb = cv2.cvtColor(img, cv2.COLOR_BGR2RGB)
    image_resized = cv2.resize(image_rgb, self.shape, interpolation=cv2.INTER_LINEAR)
    image_resized = np.transpose(image_resized, (2, 0, 1)).astype(np.float32)
    image = np.expand_dims(image_resized, axis=0)
    image /= 255.0
    
    end = time.perf_counter()
    read_time = end - start
    self.read_time += read_time

    start = time.perf_counter()
    input_name = self.session.get_inputs()[0].name
    outputs = self.session.run(None, {input_name: image})

    outputs[0] = outputs[0].reshape(1, -1, 1, 4)
    outputs[1] = outputs[1].reshape(1, -1, len(self.classes))
    embeddings = None
    if len(outputs) > 2:
      embeddings = outputs[2]
      
    embedding_path = f"{self.output_dir}/{basename}.features"

    end = time.perf_counter()
    predict_time = end - start
    self.predict_time += predict_time

    start = time.perf_counter()
    boxes = post_processing(image, 0.4, 0.6, outputs)
    end = time.perf_counter()
    process_time = end - start
    self.process_time += process_time
    
    start = time.perf_counter()
    if self.output_dir is not None:
      torch.save(embeddings, embedding_path)
    end = time.perf_counter()
    embed_time = end - start
    self.embed_time += embed_time

    start = time.perf_counter()
    if self.output_dir is not None:
      plot_boxes_cv2(img, boxes[0],
        savename=f"{self.output_dir}/{basename}", class_names=self.classes)
    write_time = time.perf_counter() - start
    self.write_time += write_time

    print(f"onnx: predict for {image_file}")
    #print(f"  output: {outputs}")
    print(f"  read_time: {read_time:.4f}s")
    print(f"  predict_time: {predict_time:.4f}s")
    print(f"  post_processing: {process_time:.4f}s")
    print(f"  write_time: {write_time:.4f}s")
    print(f"  embed_time: {embed_time:.4f}s")

  def run(self):
    """
    executes thread logic
    """
    self.load()
    
    start = time.perf_counter()
    for image in self.images:

      print(f"run: predicting on {image}")
      self.predict(image)
      
    total_time = time.perf_counter() - start
    self.fps = self.num_predicts / total_time
        
    self.avg_read_time = self.read_time / self.num_predicts
    self.avg_predict_time = self.predict_time / self.num_predicts
    self.avg_process_time = self.process_time / self.num_predicts
    self.avg_write_time = self.write_time / self.num_predicts

    print(f"thread.{self.idx}: time for {self.num_predicts} predicts")
    print(f"  total_time: {total_time:.4f}s, fps={self.fps:.2f}")
    print(f"  context_create: total: {self.context_create_time:.4f}s")
    print(f"  image_read_time: total: {self.read_time:.4f}s, avg: {self.avg_read_time:.4f}s")
    print(f"  predict_time: {self.predict_time:.4f}s, avg: {self.avg_predict_time:.4f}s")
    print(f"  process_time: {self.process_time:.4f}s, avg: {self.avg_process_time:.4f}s")
    print(f"  write_time: {self.write_time:.4f}s, avg: {self.avg_write_time:.4f}s")
      
    
def is_image(filename):
  """
  checks if filename is an image

  Args:
  filename (str): a filename to check extensions of
  Returns:
  bool: a map of arguments as defined by the parser
  """
  ext = os.path.splitext(filename)[1].lower()

  return ext == ".jpg" or ext == ".png"

def parse_args(args):
  """
  parses command line arguments

  Args:
  args (list): list of arguments
  Returns:
  dict: a map of arguments as defined by the parser
  """
  parser = argparse.ArgumentParser(
  description="predicts from an onnx model",
  add_help=True
  )
  parser.add_argument('--cpu', action='store_true',
    dest='cpu',
    help='sets a cpu-only inference mode')
  parser.add_argument('-p','--provider', '--force-provider', action='store',
    dest='provider',
    help='sets a cpu-only inference mode')
  parser.add_argument('-i','--input','--onnx', action='store',
    dest='input', default=None,
    help='the onnx model to load')
  parser.add_argument('--image', action='store',
    dest='image', default=None,
    help='the image to test the model on')
  parser.add_argument('--image-dir', action='store',
    dest='image_dir', default=None,
    help='a directory of images to test')
  parser.add_argument('--no','--no-output', action='store_true',
    dest='no_output', default=False, 
    help='do not write labeled files to disk (performance test)')
  parser.add_argument('-o','--output-dir', action='store',
    dest='output', default="labeled_images",
    help='a directory to place labeled images')
  parser.add_argument('-t','--threads', action='store',
    dest='threads', default=1, type=int,
    help='the number of threads to run')
  

  return parser.parse_args(args)

def main():
  """
  main script entry point
  """

  options = parse_args(sys.argv[1:])
  has_images = options.image is not None or options.image_dir is not None

  if options.no_output:
    print(f"onnx: not saving labeled files")
    options.output = None
  else:
    print(f"onnx: saving labeled files to {options.output}")

  if options.input is not None and has_images:
    print(f"onnx: loading {options.input}")

    if not os.path.isfile(options.input):
      print(f"predict_onnx: onnx file cannot be read. "
        "check file exists or permissions.")
      exit(1)

    basename = os.path.splitext(options.input)[0]
    names_file = f"{basename}.names"

    if options.output is not None and not os.path.isdir(options.output):
      os.makedirs(options.output)

    providers = []

    if options.provider is not None:
      providers.append(options.provider)
      providers.append('CPUExecutionProvider')
    elif not options.cpu:
      if 'CUDAExecutionProvider' in available_providers:
        providers.extend(['CUDAExecutionProvider', 'CPUExecutionProvider'])
      else:                   
        providers.extend(available_providers)
    else:
      providers.append('CPUExecutionProvider')

    if providers[0] == "ROCMExecutionProvider":
      os.environ["HIP_VISIBLE_DEVICES"]="0"


    classes = load_class_names(names_file)
    images = []

    if options.image is not None:
      images.append(options.image)

    if options.image_dir is not None:

      for dir, _, files in os.walk(options.image_dir):
        for file in files:
          source = f"{dir}/{file}"

          # file needs to be video extension and not already in cameras
          if is_image(file):
            images.append(source)

    if len(images) > 0:
      worker_threads = list()
      stats = Stats()
      
      for i in range(options.threads):
        print(f"Creating thread.{i}")
        thread = OnnxThread(i, options.input, providers, classes,
          options.output, options.threads, images)
        worker_threads.append(thread)
      
      work_start = time.perf_counter()
      
      for worker in worker_threads:
        worker.start()
        
      for worker in worker_threads:
        worker.join()
        stats.add(worker)
        
      work_time = time.perf_counter() - work_start
      
      stats.print(work_time)

  else:

    print("No model or image specified. Printing usage and help.")
    parse_args(["-h"])

if __name__ == '__main__':
  main()



# -*- coding: utf-8 -*-
'''
@Time      : 2020/05/08 11:45
@Author    : Tianxiaomo
@File      : coco_annotatin.py
@Noice     :
@Modificattion :
  @Author  :
  @Time    :
  @Detail  :

'''
import json
from collections import defaultdict
from tqdm import tqdm
import os

"""hyper parameters"""
json_file_path = 'E:/Dataset/mscoco2017/annotations/instances_train2017.json'
images_dir_path = 'mscoco2017/train2017/'
output_path = '../data/val.txt'

"""load json file"""
name_box_id = defaultdict(list)
id_name = dict()
with open(json_file_path, encoding='utf-8') as f:
  data = json.load(f)

"""generate labels"""
images = data['images']
annotations = data['annotations']
for ant in tqdm(annotations):
  id = ant['image_id']
  # name = os.path.join(images_dir_path, images[id]['file_name'])
  name = os.path.join(images_dir_path, '{:012d}.jpg'.format(id))
  cat = ant['category_id']

  if cat >= 1 and cat <= 11:
    cat = cat - 1
  elif cat >= 13 and cat <= 25:
    cat = cat - 2
  elif cat >= 27 and cat <= 28:
    cat = cat - 3
  elif cat >= 31 and cat <= 44:
    cat = cat - 5
  elif cat >= 46 and cat <= 65:
    cat = cat - 6
  elif cat == 67:
    cat = cat - 7
  elif cat == 70:
    cat = cat - 9
  elif cat >= 72 and cat <= 82:
    cat = cat - 10
  elif cat >= 84 and cat <= 90:
    cat = cat - 11

  name_box_id[name].append([ant['bbox'], cat])

"""write to txt"""
with open(output_path, 'w') as f:
  for key in tqdm(name_box_id.keys()):
    f.write(key)
    box_infos = name_box_id[key]
    for info in box_infos:
      x_min = int(info[0][0])
      y_min = int(info[0][1])
      x_max = x_min + int(info[0][2])
      y_max = y_min + int(info[0][3])

      box_info = " %d,%d,%d,%d,%d" % (
        x_min, y_min, x_max, y_max, int(info[1]))
      f.write(box_info)
    f.write('\n')




############################################################################
# --------------------------------------------------------------------------
# @author James Edmondson <james@koshee.ai>
# --------------------------------------------------------------------------
############################################################################

"""
provides utilities for tensorrt
"""

import os
import pathlib
import platform

def load_tensorrt_libs(full_search=False):
  """
  searches for tensorrt libs, starting from project directory, assuming that
  the project was installed locally or via pip. Can search whole system also,
  but that takes a long time. Sets LD_LIBRARY_PATH with results.
  
  Returns:
    tuple(str,str): location of ort_providers_tensorrt and libnvinfer
  """

  this_dir = pathlib.Path(__file__).parent.resolve() #tool
  project_dir = os.path.dirname(this_dir) #darknet2any/darknet2any
  project_dir = os.path.dirname(project_dir) #darknet2any project
  home_dir = os.path.expanduser("~")
  
  search_list = []
  
  if full_search:
    search_list = ["/"]
  else:
    search_list = [project_dir, home_dir]
  
  print("load_tensorrt_libs: search paths")
  for option in search_list:
    print(f"  {option}")
    
  ort_trt_lib = "libonnxruntime_providers_tensorrt"
  libnvinfer = "libnvinfer"
  
  uname_info = platform.uname()
  # print(f"System: {uname_info.system}")
  # print(f"Node Name: {uname_info.node}")
  # print(f"Release: {uname_info.release}")
  # print(f"Version: {uname_info.version}")
  # print(f"Machine: {uname_info.machine}")
  # print(f"Processor: {uname_info.processor}")

  if uname_info.system == "Windows":
    ort_trt_lib += ".dll"
    libnvinfer += ".dll"
  else:
    ort_trt_lib += ".so"
    libnvinfer += ".so"
    
  ort_tep_dir = None
  nvinfer_dir = None
  done = False
  
  for path in search_list:
    print(f"find_tensorrt_libs: searching for {ort_trt_lib} and {libnvinfer} in {path}")
    for dir, _, files in os.walk(path):
      for file in files:
        
        if ort_tep_dir is None and file.startswith(ort_trt_lib):
          abs_path = f"{dir}/{file}"
          print(f"find_tensorrt_libs: found {ort_trt_lib} at {abs_path}")
          os.environ["LD_LIBRARY_PATH"] = f"{dir}:" + os.environ.get(
            "LD_LIBRARY_PATH")
          ort_tep_dir = dir
          
          if ort_tep_dir is not None and nvinfer_dir is not None:
            done = True
            break
        elif nvinfer_dir is None and file.startswith(libnvinfer):
          abs_path = f"{dir}/{file}"
          print(f"find_tensorrt_libs: found {libnvinfer} at {abs_path}")
          os.environ["LD_LIBRARY_PATH"] = f"{dir}:" + os.environ.get(
            "LD_LIBRARY_PATH")
          nvinfer_dir = dir
          
          if ort_tep_dir is not None and nvinfer_dir is not None:
            done = True
            break
      
      if done:
        break
  
    if done:
      break
    
  print(f"find_tensorrt_libs: LD_LIBRARY_PATH=" + os.environ.get(
    "LD_LIBRARY_PATH"))
  
  return ort_tep_dir, nvinfer_dir

    
        
      
  



import torch.nn as nn
import torch.nn.functional as F
from typing import List

import numpy as np
from darknet2any.tool.region_loss import RegionLoss
from darknet2any.tool.yolo_layer import YoloLayer
from darknet2any.tool.config import *
from darknet2any.tool.torch_utils import *


class Mish(torch.nn.Module):
  def __init__(self):
    super().__init__()

  def forward(self, x):
    x = x * (torch.tanh(torch.nn.functional.softplus(x)))
    return x


class MaxPoolDark(nn.Module):
  def __init__(self, size=2, stride=1):
    super(MaxPoolDark, self).__init__()
    self.size = size
    self.stride = stride

  def forward(self, x):
    '''
    darknet output_size = (input_size + p - k) / s +1
    p : padding = k - 1
    k : size
    s : stride
    torch output_size = (input_size + 2*p -k) / s +1
    p : padding = k//2
    '''
    p = self.size // 2
    if ((x.shape[2] - 1) // self.stride) != ((x.shape[2] + 2 * p - self.size) // self.stride):
      padding1 = (self.size - 1) // 2
      padding2 = padding1 + 1
    else:
      padding1 = (self.size - 1) // 2
      padding2 = padding1
    if ((x.shape[3] - 1) // self.stride) != ((x.shape[3] + 2 * p - self.size) // self.stride):
      padding3 = (self.size - 1) // 2
      padding4 = padding3 + 1
    else:
      padding3 = (self.size - 1) // 2
      padding4 = padding3
    x = F.max_pool2d(F.pad(x, (padding3, padding4, padding1, padding2), mode='replicate'),
             self.size, stride=self.stride)
    return x


class Upsample_expand(nn.Module):
  def __init__(self, stride=2):
    super(Upsample_expand, self).__init__()
    self.stride = stride

  def forward(self, x):
    assert (x.data.dim() == 4)

    x = x.view(x.size(0), x.size(1), x.size(2), 1, x.size(3), 1).\
      expand(x.size(0), x.size(1), x.size(2), self.stride, x.size(3), self.stride).contiguous().\
      view(x.size(0), x.size(1), x.size(2) * self.stride, x.size(3) * self.stride)

    return x


class Upsample_interpolate(nn.Module):
  def __init__(self, stride):
    super(Upsample_interpolate, self).__init__()
    self.stride = stride

  def forward(self, x):
    assert (x.data.dim() == 4)

    out = F.interpolate(x, size=(x.size(2) * self.stride, x.size(3) * self.stride), mode='nearest')
    return out


class Reorg(nn.Module):
  def __init__(self, stride=2):
    super(Reorg, self).__init__()
    self.stride = stride

  def forward(self, x):
    stride = self.stride
    assert (x.data.dim() == 4)
    B = x.data.size(0)
    C = x.data.size(1)
    H = x.data.size(2)
    W = x.data.size(3)
    assert (H % stride == 0)
    assert (W % stride == 0)
    ws = stride
    hs = stride
    x = x.view(B, C, H / hs, hs, W / ws, ws).transpose(3, 4).contiguous()
    x = x.view(B, C, H / hs * W / ws, hs * ws).transpose(2, 3).contiguous()
    x = x.view(B, C, hs * ws, H / hs, W / ws).transpose(1, 2).contiguous()
    x = x.view(B, hs * ws * C, H / hs, W / ws)
    return x


class GlobalAvgPool2d(nn.Module):
  def __init__(self):
    super(GlobalAvgPool2d, self).__init__()

  def forward(self, x):
    N = x.data.size(0)
    C = x.data.size(1)
    H = x.data.size(2)
    W = x.data.size(3)
    x = F.avg_pool2d(x, (H, W))
    x = x.view(N, C)
    return x


# for route, shortcut and sam
class EmptyModule(nn.Module):
  def __init__(self):
    super(EmptyModule, self).__init__()

  def forward(self, x):
    return x


# support route shortcut and reorg
class Darknet(nn.Module):
  __constants__ = ['models']
  def __init__(self, cfgfile, inference=False, include_embeddings=False):
    super(Darknet, self).__init__()
    self.include_embeddings = include_embeddings
    self.inference = inference
    self.training = not self.inference

    self.blocks = parse_cfg(cfgfile)
    self.width = int(self.blocks[0]['width'])
    self.height = int(self.blocks[0]['height'])

    self.models = self.create_network(self.blocks)  # merge conv, bn,leaky
    self.loss = self.models[len(self.models) - 1]

    if self.blocks[(len(self.blocks) - 1)]['type'] == 'region':
      self.anchors = self.loss.anchors
      self.num_anchors = self.loss.num_anchors
      self.anchor_step = self.loss.anchor_step
      self.num_classes = self.loss.num_classes

    self.header = torch.IntTensor([0, 0, 0, 0])
    self.seen = 0
    self.outputs = dict()

  def forward(self, x):
    ind = -2
    self.loss = None
    self.outputs = dict()
    out_boxes = []
    for block in self.blocks:
      ind = ind + 1
      # if ind > 0:
      #  return x

      if block['type'] == 'net':
        continue
      elif block['type'] in ['convolutional', 'maxpool', 'reorg', 'upsample', 'avgpool', 'softmax', 'connected']:
        x = self.models[ind](x)
        self.outputs[ind] = x
      elif block['type'] == 'route':
        layers = block['layers'].split(',')
        layers = [int(i) if int(i) > 0 else int(i) + ind for i in layers]
        if len(layers) == 1:
          if 'groups' not in block.keys() or int(block['groups']) == 1:
            x = self.outputs[layers[0]]
            self.outputs[ind] = x
          else:
            groups = int(block['groups'])
            group_id = int(block['group_id'])
            _, b, _, _ = self.outputs[layers[0]].shape
            x = self.outputs[layers[0]][:, b // groups * group_id:b // groups * (group_id + 1)]
            self.outputs[ind] = x
        elif len(layers) == 2:
          x1 = self.outputs[layers[0]]
          x2 = self.outputs[layers[1]]
          x = torch.cat((x1, x2), 1)
          self.outputs[ind] = x
        elif len(layers) == 4:
          x1 = self.outputs[layers[0]]
          x2 = self.outputs[layers[1]]
          x3 = self.outputs[layers[2]]
          x4 = self.outputs[layers[3]]
          x = torch.cat((x1, x2, x3, x4), 1)
          self.outputs[ind] = x
        else:
          print("rounte number > 2 ,is {}".format(len(layers)))

      elif block['type'] == 'shortcut':
        from_layer = int(block['from'])
        activation = block['activation']
        from_layer = from_layer if from_layer > 0 else from_layer + ind
        x1 = self.outputs[from_layer]
        x2 = self.outputs[ind - 1]
        x = x1 + x2
        if activation == 'leaky':
          x = F.leaky_relu(x, 0.1, inplace=True)
        elif activation == 'relu':
          x = F.relu(x, inplace=True)
        self.outputs[ind] = x
      elif block['type'] == 'sam':
        from_layer = int(block['from'])
        from_layer = from_layer if from_layer > 0 else from_layer + ind
        x1 = self.outputs[from_layer]
        x2 = self.outputs[ind - 1]
        x = x1 * x2
        self.outputs[ind] = x
      elif block['type'] == 'region':
        continue
        if self.loss:
          self.loss = self.loss + self.models[ind](x)
        else:
          self.loss = self.models[ind](x)
        self.outputs[ind] = None
      elif block['type'] == 'yolo':
        # if self.training:
        #   pass
        # else:
        #   boxes = self.models[ind](x)
        #   out_boxes.append(boxes)
        boxes = self.models[ind](x)
        out_boxes.append(boxes)
      elif block['type'] == 'cost':
        continue
      else:
        print('unknown type %s' % (block['type']))

    if self.training:
      return out_boxes
    else:
      return get_region_boxes(
        out_boxes, include_embeddings=self.include_embeddings)

  def print_network(self):
    print_cfg(self.blocks)

  def get_example(self):
    return torch.rand(1, 3, self.height, self.width)

  def create_network(self, blocks):
    models = nn.ModuleList()

    prev_filters = 3
    out_filters = []
    prev_stride = 1
    out_strides = []
    conv_id = 0
    for block in blocks:
      if block['type'] == 'net':
        prev_filters = int(block['channels'])
        continue
      elif block['type'] == 'convolutional':
        conv_id = conv_id + 1
        batch_normalize = int(block['batch_normalize'])
        filters = int(block['filters'])
        kernel_size = int(block['size'])
        stride = int(block['stride'])
        is_pad = int(block['pad'])
        pad = (kernel_size - 1) // 2 if is_pad else 0
        activation = block['activation']
        model = nn.Sequential()
        if batch_normalize:
          model.add_module('conv{0}'.format(conv_id),
                   nn.Conv2d(prev_filters, filters, kernel_size, stride, pad, bias=False))
          model.add_module('bn{0}'.format(conv_id), nn.BatchNorm2d(filters))
          # model.add_module('bn{0}'.format(conv_id), BN2d(filters))
        else:
          model.add_module('conv{0}'.format(conv_id),
                   nn.Conv2d(prev_filters, filters, kernel_size, stride, pad))
        if activation == 'leaky':
          model.add_module('leaky{0}'.format(conv_id), nn.LeakyReLU(0.1, inplace=True))
        elif activation == 'relu':
          model.add_module('relu{0}'.format(conv_id), nn.ReLU(inplace=True))
        elif activation == 'mish':
          model.add_module('mish{0}'.format(conv_id), Mish())
        elif activation == 'linear':
          model.add_module('linear{0}'.format(conv_id), nn.Identity())
        elif activation == 'logistic':
          model.add_module('sigmoid{0}'.format(conv_id), nn.Sigmoid())
        else:
          print("No convolutional activation named {}".format(activation))

        prev_filters = filters
        out_filters.append(prev_filters)
        prev_stride = stride * prev_stride
        out_strides.append(prev_stride)
        models.append(model)
      elif block['type'] == 'maxpool':
        pool_size = int(block['size'])
        stride = int(block['stride'])
        if stride == 1 and pool_size % 2:
          # You can use Maxpooldark instead, here is convenient to convert onnx.
          # Example: [maxpool] size=3 stride=1
          model = nn.MaxPool2d(kernel_size=pool_size, stride=stride, padding=pool_size // 2)
        elif stride == pool_size:
          # You can use Maxpooldark instead, here is convenient to convert onnx.
          # Example: [maxpool] size=2 stride=2
          model = nn.MaxPool2d(kernel_size=pool_size, stride=stride, padding=0)
        else:
          model = MaxPoolDark(pool_size, stride)
        out_filters.append(prev_filters)
        prev_stride = stride * prev_stride
        out_strides.append(prev_stride)
        models.append(model)
      elif block['type'] == 'avgpool':
        model = GlobalAvgPool2d()
        out_filters.append(prev_filters)
        models.append(model)
      elif block['type'] == 'softmax':
        model = nn.Softmax()
        out_strides.append(prev_stride)
        out_filters.append(prev_filters)
        models.append(model)
      elif block['type'] == 'cost':
        if block['_type'] == 'sse':
          model = nn.MSELoss(reduction='mean')
        elif block['_type'] == 'L1':
          model = nn.L1Loss(reduction='mean')
        elif block['_type'] == 'smooth':
          model = nn.SmoothL1Loss(reduction='mean')
        out_filters.append(1)
        out_strides.append(prev_stride)
        models.append(model)
      elif block['type'] == 'reorg':
        stride = int(block['stride'])
        prev_filters = stride * stride * prev_filters
        out_filters.append(prev_filters)
        prev_stride = prev_stride * stride
        out_strides.append(prev_stride)
        models.append(Reorg(stride))
      elif block['type'] == 'upsample':
        stride = int(block['stride'])
        out_filters.append(prev_filters)
        prev_stride = prev_stride // stride
        out_strides.append(prev_stride)

        models.append(Upsample_expand(stride))
        # models.append(Upsample_interpolate(stride))

      elif block['type'] == 'route':
        layers = block['layers'].split(',')
        ind = len(models)
        layers = [int(i) if int(i) > 0 else int(i) + ind for i in layers]
        if len(layers) == 1:
          if 'groups' not in block.keys() or int(block['groups']) == 1:
            prev_filters = out_filters[layers[0]]
            prev_stride = out_strides[layers[0]]
          else:
            prev_filters = out_filters[layers[0]] // int(block['groups'])
            prev_stride = out_strides[layers[0]] // int(block['groups'])
        elif len(layers) == 2:
          assert (layers[0] == ind - 1 or layers[1] == ind - 1)
          prev_filters = out_filters[layers[0]] + out_filters[layers[1]]
          prev_stride = out_strides[layers[0]]
        elif len(layers) == 4:
          assert (layers[0] == ind - 1)
          prev_filters = out_filters[layers[0]] + out_filters[layers[1]] + out_filters[layers[2]] + \
                   out_filters[layers[3]]
          prev_stride = out_strides[layers[0]]
        else:
          print("route error!!!")

        out_filters.append(prev_filters)
        out_strides.append(prev_stride)
        models.append(EmptyModule())
      elif block['type'] == 'shortcut':
        ind = len(models)
        prev_filters = out_filters[ind - 1]
        out_filters.append(prev_filters)
        prev_stride = out_strides[ind - 1]
        out_strides.append(prev_stride)
        models.append(EmptyModule())
      elif block['type'] == 'sam':
        ind = len(models)
        prev_filters = out_filters[ind - 1]
        out_filters.append(prev_filters)
        prev_stride = out_strides[ind - 1]
        out_strides.append(prev_stride)
        models.append(EmptyModule())
      elif block['type'] == 'connected':
        filters = int(block['output'])
        if block['activation'] == 'linear':
          model = nn.Linear(prev_filters, filters)
        elif block['activation'] == 'leaky':
          model = nn.Sequential(
            nn.Linear(prev_filters, filters),
            nn.LeakyReLU(0.1, inplace=True))
        elif block['activation'] == 'relu':
          model = nn.Sequential(
            nn.Linear(prev_filters, filters),
            nn.ReLU(inplace=True))
        prev_filters = filters
        out_filters.append(prev_filters)
        out_strides.append(prev_stride)
        models.append(model)
      elif block['type'] == 'region':
        loss = RegionLoss()
        anchors = block['anchors'].split(',')
        loss.anchors = [float(i) for i in anchors]
        loss.num_classes = int(block['classes'])
        loss.num_anchors = int(block['num'])
        loss.anchor_step = len(loss.anchors) // loss.num_anchors
        loss.object_scale = float(block['object_scale'])
        loss.noobject_scale = float(block['noobject_scale'])
        loss.class_scale = float(block['class_scale'])
        loss.coord_scale = float(block['coord_scale'])
        out_filters.append(prev_filters)
        out_strides.append(prev_stride)
        models.append(loss)
      elif block['type'] == 'yolo':
        yolo_layer = YoloLayer()
        anchors = block['anchors'].split(',')
        anchor_mask = block['mask'].split(',')
        yolo_layer.anchor_mask = [int(i) for i in anchor_mask]
        yolo_layer.anchors = [float(i) for i in anchors]
        yolo_layer.num_classes = int(block['classes'])
        self.num_classes = yolo_layer.num_classes
        yolo_layer.num_anchors = int(block['num'])
        yolo_layer.anchor_step = len(yolo_layer.anchors) // yolo_layer.num_anchors
        yolo_layer.stride = prev_stride
        yolo_layer.scale_x_y = float(block['scale_x_y'])
        # yolo_layer.object_scale = float(block['object_scale'])
        # yolo_layer.noobject_scale = float(block['noobject_scale'])
        # yolo_layer.class_scale = float(block['class_scale'])
        # yolo_layer.coord_scale = float(block['coord_scale'])
        out_filters.append(prev_filters)
        out_strides.append(prev_stride)
        models.append(yolo_layer)
      else:
        print('unknown type %s' % (block['type']))

    return models

  def load_weights(self, weightfile):
    fp = open(weightfile, 'rb')
    header = np.fromfile(fp, count=5, dtype=np.int32)
    self.header = torch.from_numpy(header)
    self.seen = self.header[3]
    buf = np.fromfile(fp, dtype=np.float32)
    fp.close()

    start = 0
    ind = -2
    for block in self.blocks:
      if start >= buf.size:
        break
      ind = ind + 1
      if block['type'] == 'net':
        continue
      elif block['type'] == 'convolutional':
        model = self.models[ind]
        batch_normalize = int(block['batch_normalize'])
        if batch_normalize:
          start = load_conv_bn(buf, start, model[0], model[1])
        else:
          start = load_conv(buf, start, model[0])
      elif block['type'] == 'connected':
        model = self.models[ind]
        if block['activation'] != 'linear':
          start = load_fc(buf, start, model[0])
        else:
          start = load_fc(buf, start, model)
      elif block['type'] == 'maxpool':
        pass
      elif block['type'] == 'reorg':
        pass
      elif block['type'] == 'upsample':
        pass
      elif block['type'] == 'route':
        pass
      elif block['type'] == 'shortcut':
        pass
      elif block['type'] == 'sam':
        pass
      elif block['type'] == 'region':
        pass
      elif block['type'] == 'yolo':
        pass
      elif block['type'] == 'avgpool':
        pass
      elif block['type'] == 'softmax':
        pass
      elif block['type'] == 'cost':
        pass
      else:
        print('unknown type %s' % (block['type']))

  def process_tensors(self) -> torch.Tensor:
    all_parameters = []
    for module in self.models:
      for param in module.parameters():
        all_parameters.append(param.view(-1)) # Flatten each parameter tensor
    return torch.cat(all_parameters)

  def save_weights(self, outfile):
    self.cpu()
    #example = self.get_example()
    traced_model = torch.jit.script(self)
    traced_model.save(outfile)

